@@ -25,8 +25,8 @@ REST_API_CREATE_ISSUE = REST_API_BASE + '/repos/%s/%s/issues'
 REST_API_ISSUE_LABELS = REST_API_BASE + '/repos/%s/%s/issues/%s/labels'
 REST_API_COMMENTS = REST_API_BASE + '/repos/%s/%s/issues/%s/comments'
 REST_API_REACTIONS = REST_API_BASE + '/repos/%s/%s/issues/%s/reactions'
-REST_API_COMMENT = REST_API_BASE + '/repos/%s/%s/issues/comments/%s'
-REST_API_COMMENT_REACTIONS = REST_API_BASE + '/repos/%s/%s/issues/comments/%s/reactions'
+REST_API_COMMENT = REST_API_BASE + '/repos/%s/%s/%s/comments/%s'
+REST_API_COMMENT_REACTIONS = REST_API_BASE + '/repos/%s/%s/%s/comments/%s/reactions'
 REST_API_MARKDOWN = REST_API_BASE + '/markdown'
 REST_API_NOTIFICATIONS = REST_API_BASE + '/notifications?all=true&participating=true'
 GRAPHQL_BASE = 'https://api.github.com/graphql'
@@ -180,21 +180,32 @@ REACTIONS_REST = {
 REACTIONS_REST_CHARS = {char: name for name, char, in REACTIONS_REST.items()}
 
 
-# preserve <code> elements instead of converting them to `s so that GitHub
-# renders HTML entities like &gt; inside them instead of leaving them escaped.
-# background: https://chat.indieweb.org/dev/2019-12-24#t1577174464779200
-CODE_PLACEHOLDER_START = '~~~BRIDGY-CODE-TAG-START~~~'
-CODE_PLACEHOLDER_END = '~~~BRIDGY-CODE-TAG-END~~~'
+# preserve some HTML elements instead of converting them. eg <code> so that
+# GitHub renders HTML entities like &gt; inside them instead of leaving them
+# escaped. background:
+# https://chat.indieweb.org/dev/2019-12-24#t1577174464779200
+# https://github.com/snarfed/bridgy/issues/957
+PRESERVE_TAGS = ('code', 'blockquote')
 
-def tag_callback_code_placeholders(self, tag, attrs, start):
-  if tag == 'code':
-    self.o(CODE_PLACEHOLDER_START if start else CODE_PLACEHOLDER_END)
+HTTP_NON_FATAL_CODES = (
+  404,  # issue/PR or repo was probably deleted
+  410,  # ditto
+  451,  # Unavailable for Legal Reasons, eg DMCA takedown
+)
+
+
+def tag_placeholders(self, tag, attrs, start):
+  if tag in PRESERVE_TAGS:
+    self.o(f'~~~BRIDGY-{tag}-TAG-START~~~' if start else f'~~~BRIDGY-{tag}-TAG-END~~~')
     return True
 
 
-def replace_code_placeholders(text):
-  return text.replace(CODE_PLACEHOLDER_START, '<code>')\
-             .replace(CODE_PLACEHOLDER_END, '</code>')
+def replace_placeholders(text):
+  for tag in PRESERVE_TAGS:
+    text = text.replace(f'~~~BRIDGY-{tag}-TAG-START~~~', f'<{tag}>')\
+               .replace(f'~~~BRIDGY-{tag}-TAG-END~~~', f'</{tag}>')
+
+  return text
 
 
 class GitHub(source.Source):
@@ -220,8 +231,9 @@ class GitHub(source.Source):
     'ignore_links': False,
     'protect_links': False,
     'use_automatic_links': False,
-    'tag_callback': tag_callback_code_placeholders,
+    'tag_callback': tag_placeholders,
   }
+  OPTIMIZED_COMMENTS = True
 
   def __init__(self, access_token=None):
     """Constructor.
@@ -331,11 +343,8 @@ class GitHub(source.Source):
     if fetch_shares or fetch_events or fetch_mentions or search_query:
       raise NotImplementedError()
 
-    since = None
     etag_parsed = email.utils.parsedate(etag)
-    if etag_parsed:
-      since = datetime.datetime(*etag_parsed[:6])
-
+    since = datetime.datetime(*etag_parsed[:6]) if etag_parsed else None
     activities = []
 
     if activity_id:
@@ -347,7 +356,7 @@ class GitHub(source.Source):
         activities = [self.issue_to_object(issue)]
       except BaseException as e:
         code, body = util.interpret_http_exception(e)
-        if code in ('404', '410'):
+        if util.is_int(code) and int(code) in HTTP_NON_FATAL_CODES:
           activities = []
         else:
           raise
@@ -374,9 +383,9 @@ class GitHub(source.Source):
         try:
           issue = self.rest(subject_url)
         except requests.HTTPError as e:
-          if e.response.status_code in (404, 410):
+          if e.response.status_code in HTTP_NON_FATAL_CODES:
             util.interpret_http_exception(e)
-            continue  # the issue/PR or repo was (probably) deleted
+            continue
           raise
 
         obj = self.issue_to_object(issue)
@@ -435,14 +444,16 @@ class GitHub(source.Source):
 
     Returns: dict, an ActivityStreams comment object
     """
-    parts = tuple(comment_id.split(':'))
+    parts = comment_id.split(':')
     if len(parts) != 3:
       raise ValueError('GitHub comment ids must be of the form USER:REPO:COMMENT_ID')
 
-    if util.is_int(parts[2]):  # REST API id
-      comment = self.rest(REST_API_COMMENT % parts)
+    id = parts[-1]
+    if util.is_int(id):  # REST API id
+      parts.insert(2, 'issues')
+      comment = self.rest(REST_API_COMMENT % tuple(parts))
     else:  # GraphQL node id
-      comment = self.graphql(GRAPHQL_COMMENT, {'id': parts[2]})['node']
+      comment = self.graphql(GRAPHQL_COMMENT, {'id': id})['node']
 
     return self.comment_to_object(comment)
 
@@ -537,7 +548,7 @@ class GitHub(source.Source):
         abort=True,
         error_plain='You need an in-reply-to GitHub repo, issue, PR, or comment URL.')
 
-    content = orig_content = replace_code_placeholders(html.escape(
+    content = orig_content = replace_placeholders(html.escape(
       self._content_for_create(obj, ignore_formatting=ignore_formatting),
       quote=False))
     url = obj.get('url')
@@ -550,9 +561,14 @@ class GitHub(source.Source):
     if len(path) == 4:
       number = path[3]
 
-    comment_id = re.match(r'^issuecomment-([0-9]+)$', parsed.fragment)
+    # TODO: support #pullrequestreview-* URLs for top-level PR comments too.
+    # Haven't yet gotten those to work via either the issues or pulls APIs.
+    # https://github.com/snarfed/bridgy/issues/955#issuecomment-788478848
+    comment_id = re.match(r'^(discussion_r|issuecomment-)([0-9]+)$', parsed.fragment)
+    comment_type = None
     if comment_id:
-      comment_id = comment_id.group(1)
+      comment_type = 'issues' if comment_id.group(1) == 'issuecomment-' else 'pulls'
+      comment_id = comment_id.group(2)
     elif parsed.fragment:
       return source.creation_result(
         abort=True,
@@ -567,7 +583,8 @@ class GitHub(source.Source):
       is_reaction = orig_content in REACTIONS_GRAPHQL
       if preview:
         if comment_id:
-          comment = self.rest(REST_API_COMMENT % (owner, repo, comment_id))
+          comment = self.rest(REST_API_COMMENT % (owner, repo, comment_type,
+                                                  comment_id))
           target_link = '<a href="%s">a comment on %s/%s#%s, <em>%s</em></a>' % (
             base_url, owner, repo, number, util.ellipsize(comment['body']))
         else:
@@ -593,7 +610,8 @@ class GitHub(source.Source):
         # https://github.com/snarfed/bridgy/issues/824
         if is_reaction:
           if comment_id:
-            api_url = REST_API_COMMENT_REACTIONS % (owner, repo, comment_id)
+            api_url = REST_API_COMMENT_REACTIONS % (owner, repo, comment_type,
+                                                    comment_id)
             reacted = self.rest(api_url, data={
               'content': REACTIONS_REST.get(orig_content),
             })
@@ -633,14 +651,14 @@ class GitHub(source.Source):
         return source.creation_result(
           description='<span class="verb">star</span> <a href="%s">%s/%s</a>.' %
             (base_url, owner, repo))
-      else:
-        issue = self.graphql(GRAPHQL_REPO, locals())
-        resp = self.graphql(GRAPHQL_ADD_STAR, {
-          'starrable_id': issue['repository']['id'],
-        })
-        return source.creation_result({
-          'url': base_url + '/stargazers',
-        })
+
+      issue = self.graphql(GRAPHQL_REPO, locals())
+      resp = self.graphql(GRAPHQL_ADD_STAR, {
+        'starrable_id': issue['repository']['id'],
+      })
+      return source.creation_result({
+        'url': base_url + '/stargazers',
+      })
 
     elif type == 'tag':  # add label
       if not (len(path) == 4 and path[2] in ('issues', 'pull')):
@@ -666,13 +684,13 @@ class GitHub(source.Source):
         return source.creation_result(
           description='add label%s <span class="verb">%s</span> to %s.' % (
             ('s' if len(labels) > 1 else ''), ', '.join(labels), issue_link))
-      else:
-        resp = self.rest(REST_API_ISSUE_LABELS % (owner, repo, number), labels)
-        return source.creation_result({
-          'url': base_url,
-          'type': 'tag',
-          'tags': labels,
-        })
+
+      resp = self.rest(REST_API_ISSUE_LABELS % (owner, repo, number), labels)
+      return source.creation_result({
+        'url': base_url,
+        'type': 'tag',
+        'tags': labels,
+      })
 
     else:  # new issue
       if not (len(path) == 2 or (len(path) == 3 and path[2] == 'issues')):
@@ -723,7 +741,7 @@ class GitHub(source.Source):
     if not repo:
       return set()
 
-    return set(node['name'] for node in repo['labels']['nodes'])
+    return {node['name'] for node in repo['labels']['nodes']}
 
   def issue_to_object(self, issue):
     """Converts a GitHub issue or pull request to ActivityStreams.
@@ -874,7 +892,7 @@ class GitHub(source.Source):
     return self.postprocess_object(actor)
 
   def _to_object(self, input, repo_id=False):
-    """Starts to convert aHub GraphQL or REST API object to ActivityStreams.
+    """Starts to convert a GraphQL or REST API object to ActivityStreams.
 
     Args:
       input: dict, decoded JSON GraphQL or REST object
