@@ -47,6 +47,7 @@ HTML_DATA_RE = re.compile(r"""
   window\.(_sharedData\ =|__additionalDataLoaded\('[^']+',)\ *
   (.+?)
   \)?;</script>""", re.VERBOSE)
+HTML_USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:82.0) Gecko/20100101 Firefox/82.0'
 
 # URL-safe base64 encoding. used in Instagram.id_to_shortcode()
 BASE64 = string.ascii_uppercase + string.ascii_lowercase + string.digits + '-_'
@@ -71,6 +72,7 @@ class Instagram(source.Source):
   BASE_URL = 'https://www.instagram.com/'
   NAME = 'Instagram'
   FRONT_PAGE_TEMPLATE = 'templates/instagram_index.html'
+  OPTIMIZED_COMMENTS = False
 
   EMBED_POST = """
   <script async defer src="//platform.instagram.com/en_US/embeds.js"></script>
@@ -133,15 +135,15 @@ class Instagram(source.Source):
         return {}
       user_id = 'self'
 
-    if self.scrape:
-      resp = self.get_activities_response(
-        group_id=source.SELF, user_id=user_id, **kwargs)
-      items = resp.get('items')
-      return items[0].get('actor') if items else {}
-    else:
+    if not self.scrape:
       return self.user_to_actor(util.trim_nulls(
         self.urlopen(API_USER_URL % user_id) or {}))
 
+    resp = self.get_activities_response(
+      group_id=source.SELF, user_id=user_id, **kwargs)
+    items = resp.get('items')
+    return items[0].get('actor') if items else {}
+
   def get_activities_response(self, user_id=None, group_id=None, app_id=None,
                               activity_id=None, start_index=0, count=0,
                               etag=None, min_id=None, cache=None,
@@ -216,7 +218,7 @@ class Instagram(source.Source):
           cookie=cookie, fetch_extras=fetch_replies or fetch_likes, cache=cache)
       except Exception as e:
         code, body = util.interpret_http_exception(e)
-        if not ignore_rate_limit and code in ('401', '429', '503'):
+        if not ignore_rate_limit and code in ('302', '401', '429', '503'):
           logging.info('Got rate limited! Remembering for %s', str(RATE_LIMIT_BACKOFF))
           _last_rate_limited = now
           _last_rate_limited_exc = e
@@ -273,7 +275,7 @@ class Instagram(source.Source):
         body_obj = {}
 
       if body_obj.get('meta', {}).get('error_type') == 'APINotFoundError':
-        logging.warning(body_obj.get('meta', {}).get('error_message'), stack_info=True)
+        logging.warning(body_obj.get('meta', {}).get('error_message'), exc_info=True)
       else:
         raise e
 
@@ -308,12 +310,13 @@ class Instagram(source.Source):
     if cookie:
       if not cookie.startswith('sessionid='):
         cookie = 'sessionid=' + cookie
-      get_kwargs['headers'] = {'Cookie': cookie}
+      get_kwargs['headers'] = {'Cookie': cookie, 'User-Agent': HTML_USER_AGENT}
 
     resp = util.requests_get(url, **get_kwargs)
+    location = resp.headers.get('Location', '')
     if ((cookie and 'not-logged-in' in resp.text) or
         (resp.status_code in (301, 302) and
-         '/accounts/login' in resp.headers.get('Location', ''))):
+         ('/accounts/login' in location or '/challenge/' in location))):
       resp.status_code = 401
       raise requests.HTTPError('401 Unauthorized', response=resp)
     elif resp.status_code == 404:
@@ -323,7 +326,8 @@ class Instagram(source.Source):
     else:
       resp.raise_for_status()
 
-    activities, actor = self.html_to_activities(resp.text, cookie=cookie, count=count)
+    activities, actor = self.scraped_to_activities(resp.text, cookie=cookie,
+                                                   count=count)
 
     if fetch_extras:
       # batch get cached counts of comments and likes for all activities
@@ -353,7 +357,7 @@ class Instagram(source.Source):
             resp.raise_for_status()
           # otherwise resp is a fetch of just this activity; reuse it
 
-          full_activity, _ = self.html_to_activities(
+          full_activity, _ = self.scraped_to_activities(
             resp.text, cookie=cookie, count=count, fetch_extras=fetch_extras)
           if full_activity:
             activities[i] = full_activity[0]
@@ -372,13 +376,14 @@ class Instagram(source.Source):
 
     Args:
       comment_id: string comment id
-      activity_id: string activity id, required
+      activity_id: string activity id, optional
       activity_author_id: string activity author id. Ignored.
       activity: activity object, optional. Avoids fetching the activity.
     """
     if not activity:
       activity = self._get_activity(None, activity_id)
     if activity:
+      # TODO: unify with instagram, maybe in source.get_comment()
       tag_id = self.tag_uri(comment_id)
       for reply in activity.get('object', {}).get('replies', {}).get('items', []):
         if reply.get('id') == tag_id:
@@ -538,7 +543,7 @@ class Instagram(source.Source):
       'url': object.get('url'),
       'actor': object.get('author'),
       'object': object,
-      }
+    }
 
     return self.postprocess_activity(activity)
 
@@ -552,11 +557,11 @@ class Instagram(source.Source):
       an ActivityStreams object dict, ready to be JSON-encoded
     """
     id = media.get('id')
-
     user = media.get('user', {})
     content = xml.sax.saxutils.escape(media.get('caption', {}).get('text', ''))
     object = {
       'id': self.tag_uri(id),
+      'ig_shortcode': media.get('code') or media.get('shortcode'),
       # TODO: detect videos. (the type field is in the JSON respose but not
       # propagated into the Media object.)
       'objectType': OBJECT_TYPES.get(media.get('type', 'image'), 'photo'),
@@ -602,6 +607,7 @@ class Instagram(source.Source):
        for u in media.get('users_in_photo', [])] +
       [self.like_to_object(u, id, media.get('link'))
        for u in media.get('likes', {}).get('data', [])] +
+
       self._mention_tags_from_content(content)
     }
 
@@ -732,9 +738,11 @@ class Instagram(source.Source):
         'alias': '@private' if private else '@public',
       }]
 
+    pic_url = user.get('profile_picture') or user.get('profile_pic_url') or ''
+
     actor.update({
       'displayName': user.get('full_name') or username,
-      'image': {'url': user.get('profile_picture')},
+      'image': {'url': pic_url.replace(r'\/', '/')},
       'description': user.get('bio')
     })
 
@@ -780,8 +788,9 @@ class Instagram(source.Source):
 
     return ''.join(reversed(chars))
 
-  def html_to_activities(self, html, cookie=None, count=None, fetch_extras=False):
-    """Converts Instagram HTML to ActivityStreams activities.
+  def scraped_to_activities(self, html, cookie=None, count=None,
+                            fetch_extras=False):
+    """Converts scraped Instagram HTML to ActivityStreams activities.
 
     The input HTML may be from:
 
@@ -812,6 +821,7 @@ class Instagram(source.Source):
     # find media
     medias = []
     profile_user = None
+    viewer_user = None
 
     for match in matches:
       data = util.trim_nulls(json_loads(match[1]))
@@ -829,13 +839,16 @@ class Instagram(source.Source):
         edges = data['user'].get('edge_web_feed_timeline', {}).get('edges', [])
         medias.extend(e.get('node') for e in edges)
 
-      # profiles
+      # user profiles
       for page in entry_data.get('ProfilePage', []):
         profile_user = page.get('graphql', {}).get('user', {})
         medias.extend(edge['node'] for edge in
           profile_user.get('edge_owner_to_timeline_media', {}).get('edges', [])
           if edge.get('node'))
 
+      if not viewer_user:
+        viewer_user = data.get('config', {}).get('viewer')
+
       # individual photo/video permalinks
       for page in [data] + entry_data.get('PostPage', []):
         media = page.get('graphql', {}).get('shortcode_media')
@@ -859,22 +872,77 @@ class Instagram(source.Source):
 
     activities = []
     for media in util.trim_nulls(medias):
+      activity = self._json_media_node_to_activity(media, user=profile_user)
+
+      # likes
       shortcode = media.get('code') or media.get('shortcode')
       likes = media.get('edge_media_preview_like') or {}
       if shortcode and fetch_extras and likes.get('count') and not likes.get('edges'):
         # extra GraphQL fetch to get likes, as of 8/2018
-        data = self._scrape_json(HTML_LIKES_URL % shortcode, cookie=cookie)
-        likes['edges'] = data.get('data', {}).get('shortcode_media', {})\
-                             .get('edge_liked_by', {}).get('edges', [])
-      activities.append(self._json_media_node_to_activity(media, user=profile_user))
+        likes_json = self._scrape_json(HTML_LIKES_URL % shortcode, cookie=cookie)
+        self.merge_scraped_reactions(likes_json, activity)
 
-    actor = None
-    user = self._json_user_to_user(data.get('config', {}).get('viewer') or profile_user)
-    if user:
-      actor = self.user_to_actor(user)
+      activities.append(util.trim_nulls(activity))
 
+    user = self._json_user_to_user(viewer_user or profile_user)
+    actor = self.user_to_actor(user) if user else None
     return activities, actor
 
+  html_to_activities = scraped_to_activities
+
+  def scraped_to_activity(self, html, **kwargs):
+    """Converts HTML from photo/video permalink page to an AS1 activity.
+
+    Args:
+      html: string, HTML from a photo/video page on instagram.com
+      kwargs: passed through to scraped_to_activities
+
+    Returns:
+      tuple: (AS activity or None, AS logged in actor (ie viewer))
+    """
+    activities, actor = self.scraped_to_activities(html, **kwargs)
+    return (activities[0] if activities else None), actor
+
+  def scraped_to_actor(self, html, **kwargs):
+    """Extracts and returns the logged in actor from any Instagram HTML.
+
+    Args:
+      html: unicode string
+
+    Returns: dict, AS1 actor
+    """
+    return self.scraped_to_activities(html, **kwargs)[1]
+
+  def merge_scraped_reactions(self, scraped, activity):
+    """Converts and merges scraped likes and reactions into an activity.
+
+    New likes and emoji reactions are added to the activity in 'tags'.
+    Existing likes and emoji reactions in 'tags' are ignored.
+
+    Args:
+      scraped: str or dict, scraped JSON likes
+      activity: dict, AS activity to merge these reactions into
+
+    Returns:
+      list of dict AS like tag objects converted from scraped
+
+    Raises:
+      ValueError: if scraped is not valid JSON
+    """
+    if isinstance(scraped, str):
+      scraped = json_loads(scraped)
+
+    media = scraped.get('data', {}).get('shortcode_media', {})
+    if media:
+      id = util.parse_tag_uri(activity['id'])[1]
+      media_url = self.media_url(media['shortcode'])
+      likes = [self.like_to_object(like.get('node', {}), id, media_url)
+               for like in media.get('edge_liked_by', {}).get('edges', [])]
+      source.merge_by_id(activity['object'], 'tags', likes)
+      return likes
+
+    return []
+
   @staticmethod
   def _scrape_json(url, cookie=None):
     """Fetches and returns JSON from www.instagram.com."""
@@ -882,16 +950,16 @@ class Instagram(source.Source):
     if cookie:
       if not cookie.startswith('sessionid='):
         cookie = 'sessionid=' + cookie
-      headers = {'Cookie': cookie}
+      headers = {'Cookie': cookie, 'User-Agent': HTML_USER_AGENT}
 
     resp = util.requests_get(url, allow_redirects=False, headers=headers)
     resp.raise_for_status()
 
     try:
       return json_loads(resp.text)
-    except ValueError as e:
+    except ValueError:
       msg = "Couldn't decode response as JSON:\n%s" % resp.text
-      logging.error(msg, stack_info=True)
+      logging.error(msg, exc_info=True)
       resp.status_code = 504
       raise requests.HTTPError('504 Bad response from Instagram\n' + msg,
                                response=resp)
@@ -923,7 +991,7 @@ class Instagram(source.Source):
       'caption': {'text': media.get('edge_media_to_caption', {})
                                .get('edges', [{}])[0].get('node', {}).get('text', '')},
       'images': {'standard_resolution': {
-        'url': image_url.replace('\/', '/'),
+        'url': image_url.replace(r'\/', '/'),
         'width': dims.get('width'),
         'height': dims.get('height'),
       }},
@@ -937,11 +1005,21 @@ class Instagram(source.Source):
     if id and owner_id:
       media['id'] = '%s_%s' % (id, owner_id)
 
-    comments = (media.get('comments') or media.get('edge_media_to_comment') or
-                media.get('edge_media_to_parent_comment') or {})
+    comments_edge = (media.get('comments') or media.get('edge_media_to_comment') or
+                     media.get('edge_media_to_parent_comment') or {})
+    comments = [c.get('node') for c in comments_edge.get('edges', [])]
+    count = comments_edge.get('count')
+    for comment in comments:
+      threaded = comment.get('edge_threaded_comments')
+      if threaded:
+        comments += [c.get('node') for c in threaded.get('edges', [])]
+        threaded_count = threaded.get('count')
+        if threaded_count:
+          count = count + threaded_count if count else threaded_count
+
     media['comments'] = {
-      'data': [c.get('node') for c in comments.get('edges', [])],
-      'count': comments.get('count'),
+      'data': comments,
+      'count': count,
     }
 
     likes = media.get('likes') or media.get('edge_media_preview_like') or {}
@@ -953,7 +1031,8 @@ class Instagram(source.Source):
     for obj in [media] + media['comments']['data'] + media['likes']['data']:
       obj.setdefault('user', obj.get('owner') or {})
       user = obj['user'] or obj
-      user['profile_picture'] = user.get('profile_pic_url', '').replace('\/', '/')
+      if not user.get('profile_picture'):
+        user['profile_picture'] = user.get('profile_pic_url', '').replace(r'\/', '/')
 
     for c in media['comments']['data']:
       c['from'] = c['user']
@@ -963,7 +1042,7 @@ class Instagram(source.Source):
       media.update({
         'type': 'video',
         'videos': {'standard_resolution': {
-          'url': media.get('video_url', '').replace('\/', '/'),
+          'url': media.get('video_url', '').replace(r'\/', '/'),
           'width': dims.get('width'),
           'height': dims.get('height'),
         }},
@@ -1003,10 +1082,10 @@ class Instagram(source.Source):
       user = user['user']
     profile = user.get('profile_pic_url')
     if profile:
-      user['profile_picture'] = profile.replace('\/', '/')
+      user['profile_picture'] = profile.replace(r'\/', '/')
     website = user.get('external_url')
     if website:
-      user['website'] = website.replace('\/', '/')
+      user['website'] = website.replace(r'\/', '/')
     user.setdefault('bio', user.get('biography'))
 
     return user
