@@ -3,6 +3,9 @@
 
 Mastodon is an ActivityPub implementation, but it also has a REST + OAuth 2 API
 independent of AP. API docs: https://docs.joinmastodon.org/api/
+
+May also be used for services with Mastodon-compatible APIs, eg Pleroma:
+https://docs-develop.pleroma.social/backend/API/differences_in_mastoapi_responses/
 """
 import itertools
 import logging
@@ -31,11 +34,14 @@ API_STATUSES = '/api/v1/statuses'
 API_TIMELINE = '/api/v1/timelines/home'
 API_VERIFY_CREDENTIALS = '/api/v1/accounts/verify_credentials'
 
-# https://docs.joinmastodon.org/api/rest/media/#parameters
+# https://docs.joinmastodon.org/user/posting/#text
+DEFAULT_TRUNCATE_TEXT_LENGTH = 500
+
+# https://docs.joinmastodon.org/methods/statuses/media/
 MAX_ALT_LENGTH = 420
 
 # maps Mastodon media attachment type to AS1 objectType
-# https://docs.joinmastodon.org/api/entities/#type
+# https://docs.joinmastodon.org/entities/attachment/#type
 MEDIA_TYPES = {
   'image': 'image',
   'video': 'video',
@@ -50,7 +56,7 @@ MAX_MEDIA = 4
 # https://github.com/tootsuite/mastodon/blob/6bee7b820dcde6d487e93b8699d4aab3e49bedc4/app/models/account.rb#L52-L53
 USERNAME_RE = re.compile(r'[a-z0-9_]+([a-z0-9_\.-]+[a-z0-9_]+)?', re.IGNORECASE)
 MENTION_RE  = re.compile(r'(?<![\/\w])@((' + USERNAME_RE.pattern +
-                         ')(?:@[a-z0-9\.\-]+[a-z0-9]+)?)', re.IGNORECASE)
+                         r')(?:@[a-z0-9\.\-]+[a-z0-9]+)?)', re.IGNORECASE)
 
 # copied from twitter.py. if we need anything better, we could copy Mastodon's:
 # https://github.com/tootsuite/mastodon/blob/915f3712ae7ae44c0cbe50c9694c25e3ee87a540/app/models/tag.rb#L28-L30
@@ -68,12 +74,18 @@ class Mastodon(source.Source):
   DOMAIN = 'N/A'
   BASE_URL = 'N/A'
   NAME = 'Mastodon'
+  TYPE_LABELS = {
+    'post': 'toot',
+    'comment': 'reply',
+    'repost': 'boost',
+    'like': 'favorite',
+  }
 
-  # https://docs.joinmastodon.org/usage/basics/#text
-  TRUNCATE_TEXT_LENGTH = 500
+  # TRUNCATE_TEXT_LENGTH is set in the constructor
   TRUNCATE_URL_LENGTH = 23
 
-  def __init__(self, instance, access_token, user_id=None):
+  def __init__(self, instance, access_token, user_id=None,
+               truncate_text_length=None):
     """Constructor.
 
     If user_id is not provided, it will be fetched via the API.
@@ -83,11 +95,16 @@ class Mastodon(source.Source):
       user_id: string or integer, optional, current user's id (not username!) on
         this instance
       access_token: string, optional OAuth access token
+      truncate_text_length: int, optional character limit for toots, overrides
+        the default of 500
     """
     assert instance
     self.instance = self.BASE_URL = instance
     assert access_token
     self.access_token = access_token
+    self.TRUNCATE_TEXT_LENGTH = (
+      truncate_text_length if truncate_text_length is not None
+      else DEFAULT_TRUNCATE_TEXT_LENGTH)
     self.DOMAIN = util.domain_from_link(instance)
 
     if user_id:
@@ -141,11 +158,11 @@ class Mastodon(source.Source):
     return """
   <script src="%s" async="async"></script>
   <br>
-  <iframe src="%s/embed" class="mastodon-embed shadow"
+  <iframe src="%s/embed" class="%s-embed shadow"
           style="max-width: 100%%; border: 0" width="400"
           allowfullscreen="allowfullscreen">
   </iframe>
-  """ % (urllib.parse.urljoin(obj['url'], '/embed.js'), obj['url'])
+  """ % (urllib.parse.urljoin(obj['url'], '/embed.js'), obj['url'], cls.NAME.lower())
 
   def get_activities_response(self, user_id=None, group_id=None, app_id=None,
                               activity_id=None, fetch_replies=False,
@@ -158,7 +175,8 @@ class Mastodon(source.Source):
     See :meth:`Source.get_activities_response` for details.
     """
     if user_id and group_id in (source.FRIENDS, source.ALL):
-      raise ValueError("Mastodon doesn't support group_id %s with user_id" % group_id)
+      raise ValueError("%s doesn't support group_id %s with user_id" %
+                       (self.NAME, group_id))
 
     if not user_id:
       user_id = self.user_id
@@ -180,14 +198,14 @@ class Mastodon(source.Source):
         'q': search_query,
         'resolve': True,
         'offset': start_index,
-        'limit': count if count else '',
+        **params,
       }).get('statuses', [])
     else:  # eg group_id SELF
       statuses = self._get(API_ACCOUNT_STATUSES % user_id, params=params)
 
     activities = []
 
-    # batch get memcached counts of favorites and retweets for all tweets
+    # batch get memcached counts of likes, boosts, etc
     cached = {}
     if cache is not None:
       keys = itertools.product(('AMRE', 'AMF', 'AMRB'), [s['id'] for s in statuses])
@@ -228,12 +246,12 @@ class Mastodon(source.Source):
         cache_updates['AMRB ' + id] = count
 
     if fetch_mentions:
-      # https://docs.joinmastodon.org/api/rest/notifications/
+      # https://docs.joinmastodon.org/methods/notifications/
       notifs = self._get(API_NOTIFICATIONS, json={
         'exclude_types': ['follow', 'favourite', 'reblog'],
       })
-      activities.extend(self.status_to_activity(n.get('status')) for n in notifs
-                        if n.get('type') == 'mention')
+      activities.extend(self.status_to_activity(n['status']) for n in notifs
+                        if n.get('status') and n.get('type') == 'mention')
 
     resp = self.make_activities_base_response(util.trim_nulls(activities))
     if cache_updates and cache is not None:
@@ -325,7 +343,7 @@ class Mastodon(source.Source):
     }
 
     reblog = status.get('reblog')
-    base_status = reblog if reblog else status
+    base_status = reblog or status
 
     # media! into attachments.
     for media in status.get('media_attachments', []):
@@ -375,7 +393,7 @@ class Mastodon(source.Source):
       })
 
     # content: insert images for custom emoji
-    # https://docs.joinmastodon.org/api/entities/#emoji
+    # https://docs.joinmastodon.org/entities/emoji/
     content = base_status.get('content') or ''
     for emoji in base_status.get('emojis', []):
       shortcode = emoji.get('shortcode')
@@ -399,7 +417,8 @@ class Mastodon(source.Source):
     if reply_to_id:
       obj['inReplyTo'] = [{
         'id': self.tag_uri(reply_to_id),
-        'url': urllib.parse.urljoin(self.instance, '/web/statuses/' + reply_to_id),
+        # Mastodon's in_reply_to_id is str, Pixelfed's is int.
+        'url': urllib.parse.urljoin(self.instance, '/web/statuses/' + str(reply_to_id)),
       }]
 
     # to (ie visibility)
@@ -442,7 +461,12 @@ class Mastodon(source.Source):
     url = account.get('url')
     # mastodon's 'Web site' fields are HTML links, so extract their URLs
     web_sites = sum((util.extract_links(f.get('value'))
-                     for f in account.get('fields', [])), [])
+                     for f in (account.get('fields') or [])), [])
+
+    # account.created_at is string ISO8601 in Mastodon, int timestamp in Pixelfed
+    published = account.get('created_at')
+    if util.is_int(published) or util.is_float(published):
+      published = util.maybe_timestamp_to_iso8601(published)
 
     return util.trim_nulls({
       'objectType': 'person',
@@ -453,7 +477,7 @@ class Mastodon(source.Source):
       'url': url,
       'urls': [{'value': u} for u in [url] + web_sites],
       'image': {'url': account.get('avatar')},
-      'published': account.get('created_at'),
+      'published': published,
       'description': account.get('note'),
     })
 
@@ -489,7 +513,7 @@ class Mastodon(source.Source):
   def create(self, obj, include_link=source.OMIT_LINK, ignore_formatting=False):
     """Creates a status (aka toot), reply, boost (aka reblog), or favorite.
 
-    https://docs.joinmastodon.org/api/rest/statuses/
+    https://docs.joinmastodon.org/methods/statuses/
 
     Args:
       obj: ActivityStreams object
@@ -506,7 +530,7 @@ class Mastodon(source.Source):
                      ignore_formatting=False):
     """Preview creating a status (aka toot), reply, boost (aka reblog), or favorite.
 
-    https://docs.joinmastodon.org/api/rest/statuses/
+    https://docs.joinmastodon.org/methods/statuses/
 
     Args:
       obj: ActivityStreams object
@@ -523,7 +547,7 @@ class Mastodon(source.Source):
               ignore_formatting=False):
     """Creates or previews a status (aka toot), reply, boost (aka reblog), or favorite.
 
-    https://docs.joinmastodon.org/api/rest/statuses/
+    https://docs.joinmastodon.org/methods/statuses/
 
     Based on :meth:`Twitter._create`.
 
@@ -558,7 +582,6 @@ class Mastodon(source.Source):
     #
     # TODO: handle activities as well as objects? ie pull out ['object'] here if
     # necessary?
-    type = obj.get('objectType')
     prefer_content = type == 'note' or (base_url and is_reply)
     preview_description = ''
     content = self._content_for_create(
@@ -575,15 +598,16 @@ class Mastodon(source.Source):
           error_plain='No content text found.',
           error_html='No content text found.')
 
+    post_label = '%s %s' % (self.NAME, self.TYPE_LABELS['post'])
     if is_reply and not base_url:
       return source.creation_result(
         abort=True,
-        error_plain='Could not find a Mastodon toot to reply to.',
-        error_html='Could not find a Mastodon toot to <a href="http://indiewebcamp.com/reply">reply to</a>. Check that your post has the right <a href="http://indiewebcamp.com/comment">in-reply-to</a> link.')
+        error_plain='Could not find a %s to reply to.' % post_label,
+        error_html='Could not find a %s to <a href="http://indiewebcamp.com/reply">reply to</a>. Check that your post has the right <a href="http://indiewebcamp.com/comment">in-reply-to</a> link.' % post_label)
 
     # truncate and ellipsize content if necessary
     # TODO: don't count domains in remote mentions.
-    # https://docs.joinmastodon.org/usage/basics/#text
+    # https://docs.joinmastodon.org/user/posting/#text
     content = self.truncate(content, obj.get('url'), include_link, type)
 
     # linkify user mentions
@@ -607,38 +631,40 @@ class Mastodon(source.Source):
       if not base_url:
         return source.creation_result(
           abort=True,
-          error_plain='Could not find a Mastodon toot to favorite.',
-          error_html='Could not find a Mastodon toot on to <a href="http://indiewebcamp.com/favorite">favorite</a>. Check that your post has the right <a href="http://indiewebcamp.com/like">u-like-of link</a>.')
+          error_plain='Could not find a %s to %s.' % (post_label, self.TYPE_LABELS['like']),
+          error_html='Could not find a %s to <a href="http://indiewebcamp.com/like">%s</a>. Check that your post has the right <a href="http://indiewebcamp.com/like">u-like-of link</a>.' % (post_label, self.TYPE_LABELS['like']))
 
       if preview:
-        preview_description += '<span class="verb">favorite</span> <a href="%s">this toot</a>: %s' % (base_url, self.embed_post(base_obj))
+        preview_description += '<span class="verb">%s</span> <a href="%s">this %s</a>: %s' % (self.TYPE_LABELS['like'], base_url, self.TYPE_LABELS['post'], self.embed_post(base_obj))
         return source.creation_result(description=preview_description)
       else:
         resp = self._post(API_FAVORITE % base_id)
+        resp['url'] += '#favorited-by-%s' % self.user_id
         resp['type'] = 'like'
 
     elif type == 'activity' and verb == 'share':
       if not base_url:
         return source.creation_result(
           abort=True,
-          error_plain='Could not find a Mastodon toot to boost.',
-          error_html='Could not find a Mastodon toot to <a href="http://indiewebcamp.com/repost">boost</a>. Check that your post has the right <a href="http://indiewebcamp.com/repost">repost-of</a> link.')
+          error_plain='Could not find a %s to %s.' % (post_label, self.TYPE_LABELS['repost']),
+          error_html='Could not find a %s to <a href="http://indiewebcamp.com/repost">%s</a>. Check that your post has the right <a href="http://indiewebcamp.com/repost">repost-of</a> link.' % (post_label, self.TYPE_LABELS['repost']))
 
       if preview:
-          preview_description += '<span class="verb">boost</span> <a href="%s">this toot</a>: %s' % (base_url, self.embed_post(base_obj))
+          preview_description += '<span class="verb">%s</span> <a href="%s">this %s</a>: %s' % (self.TYPE_LABELS['repost'], base_url, self.TYPE_LABELS['post'], self.embed_post(base_obj))
           return source.creation_result(description=preview_description)
       else:
         resp = self._post(API_REBLOG % base_id)
         resp['type'] = 'repost'
 
-    elif type in ('note', 'article') or is_reply or is_rsvp:  # a post
+    elif (type in ('note', 'article') or is_reply or is_rsvp or
+          (type == 'activity' and verb == 'post')):  # probably a bookmark
       data = {'status': content}
 
       if is_reply:
-        preview_description += '<span class="verb">reply</span> to <a href="%s">this toot</a>: %s' % (base_url, self.embed_post(base_obj))
+        preview_description += 'add a <span class="verb">%s</span> to <a href="%s">this %s</a>: %s' % (self.TYPE_LABELS['comment'], base_url, self.TYPE_LABELS['post'], self.embed_post(base_obj))
         data['in_reply_to_id'] = base_id
       else:
-        preview_description += '<span class="verb">toot</span>:'
+        preview_description += '<span class="verb">%s</span>:' % self.TYPE_LABELS['post']
 
       num_media = len(videos) + len(images)
       if num_media > MAX_MEDIA:
@@ -704,15 +730,18 @@ class Mastodon(source.Source):
     """
     for field in ('inReplyTo', 'object', 'target'):
       for base in util.get_list(obj, field):
-        # first, check if it's on local instance
         url = util.get_url(base)
+        if not url:
+          return {}
+
+        # first, check if it's on local instance
         if url.startswith(self.instance):
           return self._postprocess_base_object(base)
 
         # nope; try mastodon's search API
         try:
           results = self._get(API_SEARCH, params={'q': url, 'resolve': True})
-        except requests.RequestException as e:
+        except requests.RequestException:
           logging.info("%s URL %s doesn't look like Mastodon:", field, url)
           continue
 
@@ -732,7 +761,8 @@ class Mastodon(source.Source):
   def upload_media(self, media):
     """Uploads one or more images or videos from web URLs.
 
-    https://docs.joinmastodon.org/api/rest/media/
+    https://docs.joinmastodon.org/methods/statuses/media/
+    https://docs.joinmastodon.org/user/posting/#attachments
 
     Args:
       media: sequence of AS image or stream objects, eg:
@@ -756,7 +786,7 @@ class Mastodon(source.Source):
       # TODO: mime type check?
       with util.requests_get(url, stream=True) as fetch:
         fetch.raise_for_status()
-        upload = self._post(API_MEDIA, files={'file': fetch.raw})
+        upload = self._post(API_MEDIA, files={'file': fetch.raw}, data=data)
 
       logging.info('Got: %s', upload)
       media_id = upload['id']
@@ -794,7 +824,7 @@ class Mastodon(source.Source):
     """Returns the current user's block list as a list of integer account ids.
 
     May make multiple API calls to fully fetch large block lists.
-    https://docs.joinmastodon.org/api/rest/blocks/#pagination
+    https://docs.joinmastodon.org/methods/accounts/blocks/
 
     Returns:
       sequence of integer Mastodon account ids on the current instance
