@@ -5,6 +5,7 @@ ActivityStreams 1 specs: http://activitystrea.ms/specs/
 """
 from collections import defaultdict
 import copy
+import html
 import itertools
 import logging
 import urllib.parse
@@ -64,6 +65,7 @@ AS_TO_MF2_TYPE = {
 }
 MF2_TO_AS_TYPE_VERB = {
   'article': ('article', None),
+  'bookmark': ('activity', 'post'),
   'event': ('event', None),
   'follow': ('activity', 'follow'),
   'invite': ('activity', 'invite'),
@@ -122,7 +124,7 @@ def get_html(val):
   if isinstance(val, dict) and val.get('html'):
     return val['html'].strip()
 
-  return get_text(val)
+  return html.escape(get_text(val), quote=False)
 
 
 def get_text(val):
@@ -225,10 +227,8 @@ def object_to_json(obj, trim_nulls=True, entry_class='h-entry',
       logging('Ignoring duration %r; expected int, got %s', duration.__class__)
       duration = None
 
-  sizes = []
   size = stream.get('size')
-  if size:
-    sizes = [str(size)]
+  sizes = [str(size)] if size else []
 
   # construct mf2!
   ret = {
@@ -276,7 +276,7 @@ def object_to_json(obj, trim_nulls=True, entry_class='h-entry',
   # content. emulate e- vs p- microformats2 parsing: e- if there are HTML tags,
   # otherwise p-.
   # https://indiewebcamp.com/note#Indieweb_whitespace_thinking
-  text = xml.sax.saxutils.unescape(primary.get('content', ''))
+  text = xml.sax.saxutils.unescape(primary.get('content') or '')
   html = render_content(primary, include_location=False,
                         synthesize_content=synthesize_content)
   if '<' in html:
@@ -343,6 +343,10 @@ def object_to_json(obj, trim_nulls=True, entry_class='h-entry',
         object_to_json(t, trim_nulls=False, entry_class='h-cite')
         for t in tags if source.object_type(t) == type]
 
+  # bookmarks
+  if obj_type == 'bookmark':
+    ret['properties']['bookmark-of'] = [primary.get('targetUrl')]
+
   # latitude & longitude
   lat = long = None
   position = ISO_6709_RE.match(primary.get('position') or '')
@@ -414,8 +418,10 @@ def json_to_object(mf2, actor=None, fetch_mf2=False):
     # TODO: remove once this is in mf2util
     # https://github.com/kylewm/mf2util/issues/18
     mf2_type = 'tag'
-  elif 'follow-of' in props: # ditto
+  elif 'follow-of' in props:  # ditto
     mf2_type = 'follow'
+  elif 'bookmark-of' in props:  # ditto
+    mf2_type = 'bookmark'
   else:
     # mf2 'photo' type is a note or article *with* a photo, but AS 'photo' type
     # *is* a photo. so, special case photo type to fall through to underlying
@@ -465,7 +471,6 @@ def json_to_object(mf2, actor=None, fetch_mf2=False):
         logging.debug('Unknown format for length or duration %r', duration)
         duration = None
 
-
   stream = None
   bytes = size_to_bytes(prop.get('size'))
   for type in 'audio', 'video':
@@ -492,6 +497,7 @@ def json_to_object(mf2, actor=None, fetch_mf2=False):
     'startTime': prop.get('start'),
     'endTime': prop.get('end'),
     'displayName': get_text(prop.get('name')),
+    'username': prop.get('nickname'),
     'summary': get_text(prop.get('summary')),
     'content': get_html(prop.get('content')),
     'url': urls[0] if urls else None,
@@ -548,6 +554,10 @@ def json_to_object(mf2, actor=None, fetch_mf2=False):
       # eliminate duplicates from redundant backcompat properties
       if t not in objects:
         objects.append(t)
+
+    objects.extend({'objectType': 'bookmark', 'targetUrl': url}
+                   for url in get_string_urls(props.get('bookmark-of', [])))
+
     obj.update({
       'object': objects[0] if len(objects) == 1 else objects,
       'actor': author,
@@ -558,6 +568,7 @@ def json_to_object(mf2, actor=None, fetch_mf2=False):
         raise NotImplementedError(
           'Combined in-reply-to and tag-of is not yet supported.')
       obj['object'] = obj.pop('tags')
+
   else:
     obj.update({
       'inReplyTo': [{'url': url} for url in in_reply_tos],
@@ -863,7 +874,8 @@ def render_content(obj, include_location=True, synthesize_content=True,
   Returns:
     string, rendered HTML
   """
-  content = obj.get('content', '')
+  obj_type = source.object_type(obj)
+  content = obj.get('content') or ''
 
   # extract tags. preserve order but de-dupe, ie don't include a tag more than
   # once.
@@ -922,6 +934,12 @@ def render_content(obj, include_location=True, synthesize_content=True,
     } for url in urls], obj)
     rendered_urls = set(urls)
 
+  # bookmarked URL
+  targetUrl = obj.get('targetUrl')
+  if obj_type == 'bookmark' and targetUrl:
+    content += '\nBookmark: %s' % util.pretty_link(
+      targetUrl, attrs={'class': 'u-bookmark-of'})
+
   # attachments, e.g. links (aka articles)
   # TODO: use oEmbed? http://oembed.com/ , http://code.google.com/p/python-oembed/
   if render_attachments:
@@ -932,7 +950,6 @@ def render_content(obj, include_location=True, synthesize_content=True,
 
   # generate share/like contexts if the activity does not have content
   # of its own
-  obj_type = source.object_type(obj)
   for as_type, verb in (
       ('favorite', 'Favorites'), ('like', 'Likes'), ('share', 'Shared')):
     if (not synthesize_content or obj_type != as_type or 'object' not in obj or
@@ -954,7 +971,7 @@ def render_content(obj, include_location=True, synthesize_content=True,
         author = target.get('author', target.get('actor', {}))
         # special case for twitter RT's
         if obj_type == 'share' and 'url' in obj and re.search(
-                '^https?://(?:www\.|mobile\.)?twitter\.com/', obj.get('url')):
+            r'^https?://(?:www\.|mobile\.)?twitter\.com/', obj.get('url')):
           content += 'RT <a href="%s">@%s</a> ' % (
             target.get('url', '#'), author.get('username'))
         else:
@@ -1014,16 +1031,17 @@ def _render_attachments(attachments, obj):
   content = ''
 
   for att in attachments:
-    name = att.get('displayName', '')
+    name = att.get('displayName') or ''
     stream = get_first(att, 'stream', {}).get('url') or ''
     image = get_first(att, 'image', {}).get('url') or ''
     open_a_tag = False
     content += '\n<p>'
 
-    if att.get('objectType') == 'video':
+    type = att.get('objectType')
+    if type == 'video':
       if stream:
         content += vid(stream, poster=image)
-    elif att.get('objectType') == 'audio':
+    elif type == 'audio':
       if stream:
         content += aud(stream)
     else:
@@ -1034,7 +1052,7 @@ def _render_attachments(attachments, obj):
       if image:
         content += '\n' + img(image, name)
 
-    if name:
+    if name and type != 'image':
       content += '\n<span class="name">%s</span>' % name
 
     if open_a_tag:
@@ -1113,7 +1131,8 @@ def tags_to_html(tags, classname, visible=True):
     for url in object_urls(tag):
       urls[url, name] = None
 
-  return ''.join('\n<a class="%s" href="%s">%s</a>' % (classname, url, name)
+  return ''.join('\n<a class="%s" %shref="%s">%s</a>' %
+                 (classname, '' if name else 'aria-hidden="true" ', url, name)
                  for url, name in sorted(urls.keys()))
 
 
@@ -1174,7 +1193,7 @@ def img(src, alt=''):
   """
   if isinstance(src, dict):
     assert not alt
-    alt = src.get('alt', '')
+    alt = src.get('alt') or ''
     src = src.get('value')
   return '<img class="u-photo" src="%s" alt=%s />' % (
       src, xml.sax.saxutils.quoteattr(alt or ''))
@@ -1265,3 +1284,48 @@ def size_to_bytes(size):
     return humanfriendly.parse_size(size)
   except humanfriendly.InvalidSize:
     logging.debug("Couldn't parse size %r", size)
+
+
+def prefix_image_urls(activity, prefix):
+  """Adds a prefix to all image url fields, eg to inject a caching proxy.
+
+  Skips any image url fields that already start with the prefix. URLs are *not*
+  URL-encode before adding the prefix. (This is currently used with
+  https://cloudimage.io/ , which doesn't URL-decode.)
+
+  Args:
+    activity: dict, AS1 activity. Modified in place.
+    prefix: string
+  """
+  a = activity
+  for elem in ([a, a.get('object'), a.get('author'), a.get('actor')] +
+               a.get('replies', {}).get('items', []) +
+               a.get('attachments', []) +
+               a.get('tags', [])):
+    if elem:
+      for img in util.get_list(elem, 'image'):
+        url = img.get('url')
+        if url and not url.startswith(prefix):
+          # Note that url isn't URL-encoded here, that's intentional.
+          # cloudimage.io doesn't decode it.
+          img['url'] = prefix + url
+      if elem is not a:
+        prefix_image_urls(elem, prefix)
+
+
+# TODO: unify
+def prefix_video_urls(activity, prefix):
+  a = activity
+  for elem in ([a, a.get('object'), a.get('author'), a.get('actor')] +
+               a.get('replies', {}).get('items', []) +
+               a.get('attachments', []) +
+               a.get('tags', [])):
+    if elem:
+      for stream in util.get_list(elem, 'stream'):
+        url = stream.get('url')
+        if url and not url.startswith(prefix):
+          # Note that url isn't URL-encoded here, that's intentional.
+          # The caching-proxy Cloudflare worker doesn't decode it.
+          stream['url'] = prefix + url
+      if elem is not a:
+        prefix_video_urls(elem, prefix)
