@@ -11,8 +11,9 @@ http://activitystrea.ms/specs/json/targeting/1.0/#anchor3
 """
 import collections
 import copy
-from html import unescape
+from html import escape, unescape
 import logging
+from operator import itemgetter
 import re
 import urllib.parse
 
@@ -22,13 +23,14 @@ import html2text
 from oauth_dropins.webutil import util
 from oauth_dropins.webutil.util import json_dumps, json_loads
 
+APP = '@app'
 ME = '@me'
 SELF = '@self'
 ALL = '@all'
 FRIENDS = '@friends'
-APP = '@app'
 SEARCH = '@search'
 BLOCKS = '@blocks'
+GROUPS = (ME, SELF, ALL, FRIENDS, SEARCH, BLOCKS)
 
 # values for create's include_link param
 OMIT_LINK = 'omit'
@@ -60,6 +62,7 @@ sources = {}
 CreationResult = collections.namedtuple('CreationResult', [
   'content', 'description', 'abort', 'error_plain', 'error_html'])
 
+
 class RateLimited(BaseException):
   """Raised when an API rate limits us, and we may have a partial result.
 
@@ -71,15 +74,17 @@ class RateLimited(BaseException):
     super(RateLimited, self).__init__(*args, **kwargs)
 
 
-def html_to_text(html, **kwargs):
+def html_to_text(html, baseurl='', **kwargs):
   """Converts string html to string text with html2text.
 
   Args:
+    baseurl: str, base URL to use when resolving relative URLs. Passed through
+      to HTML2Text().
     **kwargs: html2text options
       https://github.com/Alir3z4/html2text/blob/master/docs/usage.md#available-options
   """
   if html:
-    h = html2text.HTML2Text()
+    h = html2text.HTML2Text(baseurl=baseurl)
     h.unicode_snob = True
     h.body_width = 0  # don't wrap lines
     h.ignore_links = True
@@ -154,6 +159,21 @@ def object_type(obj):
   return type if type and type != 'activity' else obj.get('verb')
 
 
+def merge_by_id(obj, field, new):
+  """Merges new items by id into a field in an existing AS1 object.
+
+  Merges new items by id into the given field. If it exists, it must be a list.
+  Requires all existing and new items in the field to have ids.
+
+  Args:
+    obj: dict, AS1 object
+    field: str, name of field to merge new items into
+    new: sequence of AS1 dicts
+  """
+  merged = {o['id']: o for o in obj.get(field, []) + new}
+  obj[field] = sorted(merged.values(), key=itemgetter('id'))
+
+
 class SourceMeta(type):
   """Source metaclass. Registers all source classes in the sources global."""
   def __new__(meta, name, bases, class_dict):
@@ -187,12 +207,15 @@ class Source(object, metaclass=SourceMeta):
         Defaults to Twitter's limit, 280 characters as of 2019-10-12.
   * TRUNCATE_URL_LENGTH: integer, optional. Number of characters that URLs count
         for. Defaults to Twitter's, 23 as of 2019-10-12.
+  * OPTIMIZED_COMMENTS: boolean, whether :meth:`get_comment()` is optimized and
+    only fetches the requested comment. If False, :meth:`get_comment()` fetches
+    many or all of the post's comments to find the requested one.
   """
-
   POST_ID_RE = None
   HTML2TEXT_OPTIONS = {}
   TRUNCATE_TEXT_LENGTH = None
   TRUNCATE_URL_LENGTH = None
+  OPTIMIZED_COMMENTS = False
 
   def user_url(self, user_id):
     """Returns the URL for a user's profile."""
@@ -217,12 +240,12 @@ class Source(object, metaclass=SourceMeta):
     """
     return self.get_activities_response(*args, **kwargs)['items']
 
-  def get_activities_response(self, user_id=None, group_id=None, app_id=None,
-                              activity_id=None, start_index=0, count=0,
-                              etag=None, min_id=None, cache=None,
-                              fetch_replies=False, fetch_likes=False,
-                              fetch_shares=False, fetch_events=False,
-                              fetch_mentions=False, search_query=None, **kwargs):
+  def get_activities_response(
+      self, user_id=None, group_id=None, app_id=None, activity_id=None,
+      start_index=0, count=0, etag=None, min_id=None, cache=None,
+      fetch_replies=False, fetch_likes=False, fetch_shares=False,
+      fetch_events=False, fetch_mentions=False, search_query=None, scrape=False,
+      **kwargs):
     """Fetches and returns ActivityStreams activities and response details.
 
     Subclasses should override this. See :meth:`get_activities()` for an
@@ -265,7 +288,9 @@ class Source(object, metaclass=SourceMeta):
       fetch_events: boolean, whether to fetch the user's events also
       fetch_mentions: boolean, whether to fetch posts that mention the user
       search_query: string, an optional search query, only for use with
-         @search group_id
+        @search group_id
+      scrape: boolean, whether to scrape activities from HTML (etc) instead of
+        using an API. Not supported by all sources.
       kwargs: some sources accept extra kwargs. See their docs for details.
 
     Returns:
@@ -310,6 +335,64 @@ class Source(object, metaclass=SourceMeta):
       'updatedSince': False,
     }
 
+  def scraped_to_activities(self, scraped, count=None, fetch_extras=False,
+                            cookie=None):
+    """Converts scraped HTML (or JSON, etc) to AS activities.
+
+    Used for scraping data from the web instead of using an API. Useful for
+    sources with APIs that are restricted or have difficult approval processes.
+
+    Args:
+      scraped: str, scraped data from a feed of posts
+      count: integer, number of activities to return, None for all
+      fetch_extras: whether to make extra HTTP fetches to get likes, etc.
+      cookie: string, optional cookie to be used for subsequent HTTP
+        fetches, if necessary.
+
+    Returns:
+      tuple: ([AS activities], AS logged in actor (ie viewer))
+    """
+    raise NotImplementedError()
+
+  def scraped_to_activity(self, scraped):
+    """Converts scraped HTML (or JSON, etc) to a single AS activity.
+
+    Used for scraping data from the web instead of using an API. Useful for
+    sources with APIs that are restricted or have difficult approval processes.
+
+    Args:
+      scraped: str, scraped data from a single post permalink
+
+    Returns:
+      tuple: (AS activity or None, AS logged in actor (ie viewer))
+    """
+    raise NotImplementedError()
+
+  def scraped_to_actor(self, scraped):
+    """Converts HTML from a profile page to an AS1 actor.
+
+    Args:
+      html: string, HTML from a profile page
+
+    Returns: dict, AS1 actor
+    """
+    raise NotImplementedError()
+
+  def merge_scraped_reactions(self, scraped, activity):
+    """Converts and merges scraped likes and reactions into an activity.
+
+    New likes and emoji reactions are added to the activity in 'tags'.
+    Existing likes and emoji reactions in 'tags' are ignored.
+
+    Args:
+      scraped: str, HTML or JSON with likes and/or emoji reactions
+      activity: dict, AS activity to merge these reactions into
+
+    Returns:
+      list of dict AS like/react tag objects converted from scraped
+    """
+    raise NotImplementedError()
+
   def create(self, obj, include_link=OMIT_LINK, ignore_formatting=False):
     """Creates a new object: a post, comment, like, share, or RSVP.
 
@@ -579,7 +662,7 @@ class Source(object, metaclass=SourceMeta):
           activity['title'] = obj_name
         elif verb and (obj_name or obj_type):
           app = activity.get('generator', {}).get('displayName')
-          name = obj_name if obj_name else 'a %s' % (obj_type or 'unknown')
+          name = obj_name or 'a %s' % (obj_type or 'unknown')
           app = ' on %s' % app if app else ''
           activity['title'] = '%s %s %s%s.' % (actor_name, verb or 'posted',
                                                name, app)
@@ -608,8 +691,9 @@ class Source(object, metaclass=SourceMeta):
   _PERMASHORTCITATION_RE = re.compile(r'\(([^:\s)]+\.[^\s)]{2,})[ /]([^\s)]+)\)$')
 
   @staticmethod
-  def original_post_discovery(activity, domains=None, cache=None,
-                              include_redirect_sources=True, **kwargs):
+  def original_post_discovery(
+      activity, domains=None, cache=None, include_redirect_sources=True,
+      include_reserved_hosts=True, **kwargs):
     """Discovers original post links.
 
     This is a variation on http://indiewebcamp.com/original-post-discovery . It
@@ -629,6 +713,9 @@ class Source(object, metaclass=SourceMeta):
         (Permashortcitations are exempt.)
       include_redirect_sources: boolean, whether to include URLs that redirect
         as well as their final destination URLs
+      include_reserved_hosts: boolean, whether to include domains on reserved
+        TLDs (eg foo.example) and local hosts (eg http://foo.local/,
+        http://my-server/)
       cache: deprecated, unused
       kwargs: passed to requests.head() when following redirects
 
@@ -641,7 +728,9 @@ class Source(object, metaclass=SourceMeta):
     # find all candidate URLs
     tags = [t.get('url') for t in obj.get('attachments', []) + obj.get('tags', [])
             if t.get('objectType') in ('article', 'mention', 'note', None)]
-    candidates = tags + util.extract_links(content) + obj.get('upstreamDuplicates', [])
+    candidates = (tags + util.extract_links(content) +
+                  obj.get('upstreamDuplicates', []) +
+                  util.get_list(obj, 'targetUrl'))
 
     # Permashortcitations (http://indiewebcamp.com/permashortcitation) are short
     # references to canonical copies of a given (usually syndicated) post, of
@@ -651,8 +740,9 @@ class Source(object, metaclass=SourceMeta):
 
     candidates = set(util.dedupe_urls(
       util.clean_url(url) for url in candidates
+      if url and (url.startswith('http://') or url.startswith('https://')) and
       # heuristic: ellipsized URLs are probably incomplete, so omit them.
-      if url and not url.endswith('...') and not url.endswith('…')))
+      not url.endswith('...') and not url.endswith('…')))
 
     # check for redirect and add their final urls
     redirects = {}  # maps final URL to original URL for redirects
@@ -672,7 +762,16 @@ class Source(object, metaclass=SourceMeta):
         # this is a redirected original URL. postpone and handle it when we hit
         # its final URL so that we know the final domain.
         continue
+
       domain = util.domain_from_link(url)
+      if not domain:
+        continue
+
+      if not include_reserved_hosts and (
+          ('.' not in domain
+           or domain.split('.')[-1] in (util.RESERVED_TLDS | util.LOCAL_TLDS))):
+        continue
+
       which = (originals if not domains or util.domain_or_parent_in(domain, domains)
                else mentions)
       which.add(url)
@@ -801,7 +900,22 @@ class Source(object, metaclass=SourceMeta):
     return any(changed(before, after, field, 'activity') or
                changed(obj_b, obj_a, field, 'activity[object]')
                for field in ('objectType', 'verb', 'to', 'content', 'location',
-                             'image'))
+                             'image', 'inReplyTo'))
+
+  @staticmethod
+  def append_in_reply_to(before, after):
+    """appends the inReplyTos from the before object to the after object, in place
+
+    Args:
+      before, after: dicts, ActivityStreams activities or objects
+    """
+    obj_b = before.get('object', before)
+    obj_a = after.get('object', after)
+
+    if obj_b and obj_a:
+      reply_b = util.get_list(obj_b, 'inReplyTo')
+      reply_a = util.get_list(obj_a, 'inReplyTo')
+      obj_a['inReplyTo'] = util.dedupe_urls(reply_a + reply_b)
 
   @classmethod
   def embed_post(cls, obj):
@@ -811,16 +925,15 @@ class Source(object, metaclass=SourceMeta):
       obj: AS1 dict with at least url, and optionally also content.
 
     Returns: string, HTML
-
-    Raises: ValueError, if obj['content'] contains raw < or > characters.
     """
     obj = copy.copy(obj)
     for field in 'url', 'content':
       if field not in obj:
         obj.setdefault(field, obj.get('object', {}).get(field, ''))
 
-    if '<' in obj['content'] or '>' in obj['content']:
-      raise ValueError("obj['content'] has unescaped < or > characters!")
+    # allow HTML in posts to be rendered safely, avoiding risk of XSS, but
+    # allowing posts that could contain HTML to be allowed
+    obj['content'] = escape(obj['content'])
 
     # escape URL, but not with urllib.parse.quote, because it quotes a ton of
     # chars we want to pass through, including most unicode chars.
@@ -834,9 +947,9 @@ class Source(object, metaclass=SourceMeta):
     return """
 <a class="h-card" href="%s">
  <img class="profile u-photo" src="%s" width="32px" /> %s</a>""" % (
-   actor.get('url'),
-   actor.get('image', {}).get('url'),
-   actor.get('displayName'))
+     actor.get('url'),
+     actor.get('image', {}).get('url'),
+     actor.get('displayName'))
 
   def tag_uri(self, name):
     """Returns a tag URI string for this source and the given string name."""
@@ -973,14 +1086,14 @@ class Source(object, metaclass=SourceMeta):
     is_html = (bool(BeautifulSoup(content, 'html.parser').find()) or
                HTML_ENTITY_RE.search(content))
     if is_html and not ignore_formatting:
-      content = html_to_text(content, **self.HTML2TEXT_OPTIONS)
+      content = html_to_text(content, baseurl=(obj.get('url') or ''),
+                             **self.HTML2TEXT_OPTIONS)
     elif not is_html and ignore_formatting:
       content = re.sub(r'\s+', ' ', content)
 
-    return summary or (
-            (name or content) if prefer_name else
-            (content or name)
-           ) or ''
+    return summary or ((name or content) if prefer_name else
+                       (content or name)
+                       ) or ''
 
   def truncate(self, content, url, include_link, type=None, quote_url=None):
     """Shorten text content to fit within a character limit.
@@ -1025,4 +1138,3 @@ class Source(object, metaclass=SourceMeta):
       truncated += ' ' + quote_url
 
     return truncated
-
