@@ -107,23 +107,24 @@ class AtomTest(testutil.TestCase):
 
   def test_activities_to_atom(self):
     for test_module in test_facebook, test_instagram, test_twitter:
-      request_url = 'http://request/url?access_token=foo'
-      host_url = 'http://host/url'
-      base_url = 'http://base/url'
-      self.assert_multiline_equals(
-        test_module.ATOM % {
-          'request_url': request_url,
-          'host_url': host_url,
-          'base_url': base_url,
-        },
-        atom.activities_to_atom(
-          [copy.deepcopy(test_module.ACTIVITY)],
-          test_module.ACTOR,
-          request_url=request_url,
-          host_url=host_url,
-          xml_base=base_url
-        ),
-        ignore_blanks=True)
+      with self.subTest(test_module):
+        request_url = 'http://request/url?access_token=foo'
+        host_url = 'http://host/url'
+        base_url = 'http://base/url'
+        self.assert_multiline_equals(
+          test_module.ATOM % {
+            'request_url': request_url,
+            'host_url': host_url,
+            'base_url': base_url,
+          },
+          atom.activities_to_atom(
+            [copy.deepcopy(test_module.ACTIVITY)],
+            test_module.ACTOR,
+            request_url=request_url,
+            host_url=host_url,
+            xml_base=base_url
+          ),
+          ignore_blanks=True)
 
   def test_activity_to_atom(self):
     self.assert_multiline_equals(
@@ -910,3 +911,19 @@ going to Homebrew Website Club
     obj['actor']['image'] = obj['actor']['image'][0]
     obj['attachments'][0]['image'] = obj['attachments'][0]['image'][0]
     check()
+
+  def test_summary(self):
+    """inReplyTo should be translated to thr:in-reply-to."""
+    activity = {'object': {
+      'summary': 'foo &',
+      'content': 'bar',
+    }}
+
+    self.assert_multiline_in("""\
+<summary>foo &amp;</summary>
+<content type="xhtml">
+<div xmlns="http://www.w3.org/1999/xhtml">
+bar
+</div>
+</content>
+""", atom.activities_to_atom([activity], {}), ignore_blanks=True)
