@@ -321,7 +321,7 @@ PHOTO_COMMENTS = {
     'photo_id': '5227922370',
     'comment': [
       {
-        'id': '4942564-5227922370-72157625845945286',
+        'id': '4942564',
         'author': '36398523@N00',
         'authorname': 'if winter ends',
         'iconserver': '108',
@@ -348,7 +348,7 @@ COMMENT_OBJS = [{  # ActivityStreams
     'url': 'https://www.flickr.com/people/if_winter_ends/',
   },
   'content': 'Love this!',
-  'id': tag_uri('4942564-5227922370-72157625845945286'),
+  'id': tag_uri('4942564'),
   'updated': '2011-01-17T18:24:03+00:00',
   'published': '2011-01-17T18:24:03+00:00',
   'url': 'https://www.flickr.com/photos/kindofblue115/5227922370/#comment72157625845945286',
@@ -644,6 +644,18 @@ LIKE_OBJ = {
   'verb': 'like',
 }
 
+TAG_REPLY_ACTIVITY = {
+  'objectType': 'activity',
+  'verb': 'tag',
+  'object': [
+    {'displayName': 'foo'},
+    {'displayName': 'bar'},
+  ],
+  'target': {
+    'url': 'https://www.flickr.com/photos/marietta_wood_works/21904325000/in/contacts/',
+  },
+}
+
 # uploads send oauth params along with the post data; it's useful to
 # be able to check that they exist, but ignore their values.
 IGNORED_OAUTH_PARAMS = [
@@ -815,7 +827,12 @@ class FlickrTest(testutil.TestCase):
       }, json_dumps(PHOTO_COMMENTS))
     self.mox.ReplayAll()
     self.assert_equals(COMMENT_OBJS[0], self.flickr.get_comment(
-      '4942564-5227922370-72157625845945286', '5227922370'))
+      '4942564', '5227922370'))
+
+  def test_get_comment_with_activity(self):
+    # skips API call
+    got = self.flickr.get_comment('4942564', activity=ACTIVITY_WITH_COMMENTS)
+    self.assert_equals(COMMENT_OBJS[0], got)
 
   def test_search_raises_not_implemented(self):
     with self.assertRaises(NotImplementedError):
@@ -1126,6 +1143,23 @@ class FlickrTest(testutil.TestCase):
       },
     }, activity['object']['author'])
 
+  def test_create_and_preview_tag_reply(self):
+    preview = self.flickr.preview_create(TAG_REPLY_ACTIVITY)
+    self.assertIn('add the tags <em>bar</em>, <em>foo</em>', preview.description)
+
+    self.expect_call_api_method(
+      'flickr.photos.addTags', {'photo_id': '21904325000', 'tags': 'bar foo'}, '')
+    self.expect_call_api_method(
+      'flickr.people.getLimits', {},
+      json_dumps({'person': {'nsid': '39216764@N00'}}))
+    self.mox.ReplayAll()
+
+    self.assertEqual({
+      'type': 'tag',
+      'url': 'https://www.flickr.com/photos/marietta_wood_works/21904325000/in/contacts/#tagged-by-39216764@N00',
+      'tags': ['bar', 'foo'],
+    }, self.flickr.create(TAG_REPLY_ACTIVITY).content)
+
   def test_delete(self):
     self.expect_call_api_method('flickr.photos.delete', {'photo_id': '789'}, '')
     self.expect_call_api_method(
