@@ -201,7 +201,6 @@ PULL_REST = {  # GitHub
   'user': USER_REST,
   'title': 'a PR to merge',
   'body': 'a PR message',
-  'comments_url': 'https://api.github.com/repos/foo/bar/issues/444/comments',
   'issue_url': 'https://api.github.com/repos/foo/bar/issues/444',
   'diff_url': 'https://github.com/foo/bar/pull/444.diff',
   'patch_url': 'https://github.com/foo/bar/pull/444.patch',
@@ -274,7 +273,6 @@ COMMENT_OBJ = {  # ActivityStreams
   'url': 'https://github.com/foo/bar/pull/123#issuecomment-456',
   'author': ACTOR,
   'content': 'i have something to say here',
-  'published': '2012-12-05T00:58:26+00:00',
   'inReplyTo': [{'url': 'https://github.com/foo/bar/pull/123'}],
   'published': '2015-07-23T18:47:58+00:00',
   'updated': '2015-07-23T19:47:58+00:00',
@@ -559,6 +557,9 @@ class GitHubTest(testutil.TestCase):
   def test_get_activities_activity_id_deleted(self):
     self._test_get_activities_activity_id_fails(410)
 
+  def test_get_activities_activity_id_unavailable_for_legal_reasons(self):
+    self._test_get_activities_activity_id_fails(451)
+
   def _test_get_activities_activity_id_fails(self, status):
     self.expect_rest(REST_API_ISSUE % ('a', 'b', 1), {
       'message': 'Not Found',
@@ -578,6 +579,9 @@ class GitHubTest(testutil.TestCase):
   def test_get_activities_pr_deleted(self):
     self._test_get_activities_pr_fails(410)
 
+  def test_get_activities_pr_unavailable_for_legal_reasons(self):
+    self._test_get_activities_pr_fails(451)
+
   def _test_get_activities_pr_fails(self, status):
     self.expect_rest(REST_API_NOTIFICATIONS,
                      [NOTIFICATION_PULL_REST, NOTIFICATION_ISSUE_REST])
@@ -627,7 +631,7 @@ class GitHubTest(testutil.TestCase):
     self.assert_equals({}, self.gh.issue_to_object({}))
 
   def test_get_comment_rest(self):
-    self.expect_rest(REST_API_COMMENT % ('foo', 'bar', 123), COMMENT_REST)
+    self.expect_rest(REST_API_COMMENT % ('foo', 'bar', 'issues', 123), COMMENT_REST)
     self.mox.ReplayAll()
     self.assert_equals(COMMENT_OBJ, self.gh.get_comment('foo:bar:123'))
 
@@ -781,6 +785,28 @@ class GitHubTest(testutil.TestCase):
       'url': 'https://github.com/foo/bar/issues/123',
     }, result.content)
 
+  def test_create_with_relative_image_and_link(self):
+    self.expect_graphql_get_labels([])
+    self.expect_requests_post(github.REST_API_CREATE_ISSUE % ('foo', 'bar'), json={
+        'title': 'an issue title',
+        'body': '[foo](http://site/post/foo) ![](http://site/bar/baz)',
+        'labels': [],
+      }, response={
+        'html_url': 'https://github.com/foo/bar/issues/123',
+      }, headers=EXPECTED_HEADERS)
+    self.mox.ReplayAll()
+
+    result = self.gh.create({
+      'url': 'http://site/post/xyz',
+      'title': 'an issue title',
+      'content': """
+<a href="foo">foo</a>
+<img src="/bar/baz" />
+""",
+      'inReplyTo': [{'url': 'https://github.com/foo/bar/issues'}],
+    })
+    self.assertIsNone(result.error_plain, result)
+
   def test_create_escape_html(self):
     content = 'x &lt;data foo&gt; &amp; y'
 
@@ -821,8 +847,27 @@ class GitHubTest(testutil.TestCase):
     })
     self.assertIsNone(result.error_plain, result)
 
+  def test_create_blockquote(self):
+    content = 'x <blockquote>y</blockquote>'
+    self.expect_graphql_get_labels([])
+    self.expect_requests_post(github.REST_API_CREATE_ISSUE % ('foo', 'bar'), json={
+        'title': 'an issue title',
+        'body': content,
+        'labels': [],
+      }, response={
+        'html_url': 'https://github.com/foo/bar/issues/123',
+      }, headers=EXPECTED_HEADERS)
+    self.mox.ReplayAll()
+
+    result = self.gh.create({
+      'title': 'an issue title',
+      'content': content,
+      'inReplyTo': [{'url': 'https://github.com/foo/bar/issues'}],
+    })
+    self.assertIsNone(result.error_plain, result)
+
   def test_preview_issue(self):
-    for i in range(2):
+    for _ in range(2):
       self.expect_graphql_get_labels(['new silo'])
       rendered = self.expect_markdown_render(ISSUE_OBJ['content'].strip())
     self.mox.ReplayAll()
@@ -837,6 +882,21 @@ class GitHubTest(testutil.TestCase):
         '<span class="verb">create a new issue</span> on <a href="%s">foo/bar</a> and attempt to add label <span class="verb">new silo</span>:' % url,
         preview.description, preview)
 
+  def test_preview_blockquote(self):
+    content = 'x <blockquote>y</blockquote>'
+    self.expect_graphql_get_labels(['new silo'])
+    rendered = self.expect_markdown_render(content.strip())
+    self.mox.ReplayAll()
+
+    obj = copy.deepcopy(ISSUE_OBJ)
+    preview = self.gh.preview_create({
+      'title': 'an issue title',
+      'content': content,
+      'inReplyTo': [{'url': 'https://github.com/foo/bar/issues'}],
+    })
+    self.assertIsNone(preview.error_plain, preview)
+    self.assertEqual('<b>an issue title</b><hr>' + rendered, preview.content)
+
   def test_create_issue_private_repo(self):
     """eg the w3c/AB repo is private and returns repository: None
     https://console.cloud.google.com/errors/CMbUj5KyrvH69gE
@@ -1004,9 +1064,10 @@ class GitHubTest(testutil.TestCase):
     preview = self.gh.preview_create(REACTION_OBJ_INPUT)
     self.assertEqual(u'<span class="verb">react 👍</span> to <a href="https://github.com/foo/bar/pull/123">foo/bar#123, <em>an issue title</em></a>.', preview.description)
 
-  def test_create_reaction_comment(self):
+  def test_create_reaction_issue_comment(self):
     self.expect_requests_post(
-      REST_API_COMMENT_REACTIONS % ('foo', 'bar', 456), headers=EXPECTED_HEADERS,
+      REST_API_COMMENT_REACTIONS % ('foo', 'bar', 'issues', 456),
+      headers=EXPECTED_HEADERS,
       json={
         'content': '+1',
       }, response={
@@ -1023,8 +1084,8 @@ class GitHubTest(testutil.TestCase):
       'type': 'react',
     }, result.content, result)
 
-  def test_preview_reaction_comment(self):
-    self.expect_rest(REST_API_COMMENT % ('foo', 'bar', 456), COMMENT_REST)
+  def test_preview_reaction_issue_comment(self):
+    self.expect_rest(REST_API_COMMENT % ('foo', 'bar', 'issues', 456), COMMENT_REST)
     self.mox.ReplayAll()
 
     preview = self.gh.preview_create(COMMENT_REACTION_OBJ_INPUT)
