@@ -509,8 +509,8 @@ HTML_PHOTO_LIKES_RESPONSE = {
   'status': 'ok',
   'data': {
     'shortcode_media': {
-      'id': '1848262920796141768',
-      'shortcode': 'BmmWVV9lHjI',
+      'id': '123',
+      'shortcode': 'ABC123',
       'edge_liked_by': {
         'count': 9,
         'edges': [{
@@ -569,17 +569,49 @@ HTML_VIDEO_FULL = {
       'id': '789',
       'created_at': 1349588757,
       'text': '太可爱了。cute，@a_person, very cute',
+      'edge_threaded_comments': {
+        'count': 1,
+        'edges': [{
+          'node': {
+            'id': '020',
+            'text': 'hah, i have no tips whatsoever',
+            'created_at': 1594392712,
+            'owner': {
+              'id': '420973239',
+              'profile_pic_url': 'http:\/\/picture\/commenter/2',
+              'username': 'someone',
+            },
+          }
+        }],
+      },
     }}],
-    'count': 1,
+    'count':  1,
   },
-  'edge_media_to_tagged_user': {
-    'edges': [{'node': {
-      'user': {'username': 'ap'},
-      'x': 0.4657777507,
-      'y': 0.4284444173},
+  'edge_media_to_tagged_user':  {
+    'edges':  [{'node':  {
+      'user':  {'username':  'ap'},
+      'x':  0.4657777507,
+      'y':  0.4284444173},
     }],
   },
-  'taken_at_timestamp': 1453036552,
+  'taken_at_timestamp':  1453036552,
+}
+HTML_VIDEO_EXTRA_COMMENT_OBJ = {  # ActivityStreams
+  'objectType': 'comment',
+  'author': {
+    'objectType': 'person',
+    'id': tag_uri('420973239'),
+    'displayName': 'someone',
+    'username': 'someone',
+    'image': {'url': 'http://picture/commenter/2'},
+    'url': 'https://www.instagram.com/someone/',
+  },
+  'content': 'hah, i have no tips whatsoever',
+  'id': tag_uri('020'),
+  'published': '2020-07-10T14:51:52+00:00',
+  'url': 'https://www.instagram.com/p/XYZ789/#comment-020',
+  'inReplyTo': [{'id': tag_uri('789_456')}],
+  'to': [{'objectType':'group', 'alias':'@public'}],
 }
 
 HTML_VIDEO = copy.deepcopy(HTML_VIDEO_FULL)
@@ -635,20 +667,21 @@ HTML_SUGGESTED_USERS = {
   '__typename': 'GraphSuggestedUserFeedUnit',
 }
 
+HTML_VIEWER_CONFIG = {
+  'csrf_token': '...',
+  'viewer': {
+    'external_url': 'https:\/\/snarfed.org',
+    'biography': 'something or other',
+    'id': '420973239',
+    'full_name': 'Ryan B',
+    'profile_pic_url': 'https:\/\/scontent-sjc2-1.cdninstagram.com\/hphotos-xfa1\/t51.2885-19\/11373714_959073410822287_2004790583_a.jpg',
+    'has_profile_pic': True,
+    'username': 'snarfed',
+  },
+}
 HTML_FEED = {  # eg https://www.instagram.com/ when you're logged in
   'environment_switcher_visible_server_guess': True,
-  'config': {
-    'csrf_token': '...',
-    'viewer': {
-      'external_url': 'https:\/\/snarfed.org',
-      'biography': 'something or other',
-      'id': '420973239',
-      'full_name': 'Ryan B',
-      'profile_pic_url': 'https:\/\/scontent-sjc2-1.cdninstagram.com\/hphotos-xfa1\/t51.2885-19\/11373714_959073410822287_2004790583_a.jpg',
-      'has_profile_pic': True,
-      'username': 'snarfed',
-    }
-  },
+  'config': HTML_VIEWER_CONFIG,
   'display_properties_server_guess': {'pixel_ratio': 2.0, 'viewport_width': 1280},
   'qe': {'su': {'g': 'control', 'p': {'enabled': 'false'}}},
   'hostname': 'www.instagram.com',
@@ -845,6 +878,7 @@ HTML_PHOTO_ACTIVITY = {  # ActivityStreams
     'author': HTML_ACTOR,
     'content': 'Elvis hits out of RCA Studio B',
     'id': tag_uri('123_456'),
+    'ig_shortcode': 'ABC123',
     'published': '2016-01-17T20:46:33+00:00',
     'url': 'https://www.instagram.com/p/ABC123/',
     'image': {'url': 'https://scontent-sjc2-1.cdninstagram.com/hphotos-xfp1/t51.2885-15/e35/12545499_1662965520652470_1466520818_n.jpg'},
@@ -884,6 +918,7 @@ HTML_VIDEO_ACTIVITY = {  # ActivityStreams
     'content': 'Eye of deer \ud83d\udc41 and #selfie from me',
     'content': 'Eye of deer 👁 and #selfie from me',
     'id': tag_uri('789_456'),
+    'ig_shortcode': 'XYZ789',
     'published': '2016-01-17T13:15:52+00:00',
     'url': 'https://www.instagram.com/p/XYZ789/',
     'image': {'url': 'https://scontent-sjc2-1.cdninstagram.com/hphotos-xpf1/t51.2885-15/s750x750/sh0.08/e35/12424348_567037233461060_1986731502_n.jpg'},
@@ -916,8 +951,8 @@ HTML_VIDEO_ACTIVITY = {  # ActivityStreams
 }
 HTML_VIDEO_ACTIVITY_FULL = copy.deepcopy(HTML_VIDEO_ACTIVITY)
 HTML_VIDEO_ACTIVITY_FULL['object']['replies'] = {
-  'items': copy.deepcopy(COMMENT_OBJS),
-  'totalItems': len(COMMENT_OBJS),
+  'items': copy.deepcopy(COMMENT_OBJS) + [HTML_VIDEO_EXTRA_COMMENT_OBJ],
+  'totalItems': len(COMMENT_OBJS) + 1,
 }
 HTML_VIDEO_ACTIVITY_FULL['object']['replies']['items'][0].update({
   'url': 'https://www.instagram.com/p/XYZ789/#comment-789',
@@ -1132,7 +1167,7 @@ class InstagramTest(testutil.TestCase):
     self.mox.ReplayAll()
 
     cache = util.CacheDict()
-    for i in range(3):
+    for _ in range(3):
       self.instagram.get_activities(user_id='x', group_id=source.SELF,
                                     fetch_likes=True, fetch_replies=True,
                                     scrape=True, cache=cache, cookie='kuky')
@@ -1253,7 +1288,7 @@ class InstagramTest(testutil.TestCase):
     # first attempt: rate limited
     self.expect_requests_get(HTML_BASE_URL, status_code=429, cookie='kuky')
     # third attempt ignore rate limit lock. fourth limit is past lock.
-    for i in range(2):
+    for _ in range(2):
       self.expect_requests_get('x/', HTML_PROFILE_COMPLETE)
     self.mox.ReplayAll()
 
@@ -1305,7 +1340,7 @@ class InstagramTest(testutil.TestCase):
 
     ig = Instagram(scrape=True)
     self.assert_equals(HTML_VIDEO_ACTIVITY_FULL['object']['replies']['items'][0],
-                          ig.get_comment('789', activity_id='1208909509631101904_942513'))
+                       ig.get_comment('789', activity_id='1208909509631101904_942513'))
 
   def test_get_comment_with_activity(self):
     # skips API call
@@ -1520,36 +1555,36 @@ class InstagramTest(testutil.TestCase):
         'object': {'url': 'https://www.instagram.com/p/zHA5BLo1Mo/'},
       }))
 
-  def test_html_to_activities_feed(self):
-    activities, viewer = self.instagram.html_to_activities(HTML_FEED_COMPLETE)
+  def test_scraped_to_activities_feed(self):
+    activities, viewer = self.instagram.scraped_to_activities(HTML_FEED_COMPLETE)
     self.assert_equals(HTML_ACTIVITIES_FULL, activities)
     self.assert_equals(HTML_VIEWER, viewer)
 
-    activities, viewer = self.instagram.html_to_activities(HTML_FEED_COMPLETE_2)
+    activities, viewer = self.instagram.scraped_to_activities(HTML_FEED_COMPLETE_2)
     self.assert_equals(HTML_ACTIVITIES_FULL, activities)
 
-  def test_html_to_activities_profile(self):
-    activities, viewer = self.instagram.html_to_activities(HTML_PROFILE_COMPLETE)
+  def test_scraped_to_activities_profile(self):
+    activities, viewer = self.instagram.scraped_to_activities(HTML_PROFILE_COMPLETE)
     self.assert_equals(HTML_ACTIVITIES, activities)
     self.assert_equals(HTML_VIEWER_PUBLIC, viewer)
 
-  def test_html_to_activities_profile_private(self):
-    _, actor = self.instagram.html_to_activities(HTML_PROFILE_PRIVATE_COMPLETE)
+  def test_scraped_to_activities_profile_private(self):
+    _, actor = self.instagram.scraped_to_activities(HTML_PROFILE_PRIVATE_COMPLETE)
     self.assert_equals([{'objectType':'group', 'alias':'@private'}], actor['to'])
 
-  def test_html_to_activities_profile_fill_in_owner(self):
+  def test_scraped_to_activities_profile_fill_in_owner(self):
     profile = copy.deepcopy(HTML_PROFILE)
     user = profile['entry_data']['ProfilePage'][0]['graphql']['user']
     user['edge_owner_to_timeline_media']['edges'][0]['node']['owner'] = {
       'id': user['id'],
     }
 
-    activities, _ = self.instagram.html_to_activities(
+    activities, _ = self.instagram.scraped_to_activities(
       HTML_HEADER + json_dumps(profile) + HTML_FOOTER)
     self.assertEqual(HTML_VIEWER_PUBLIC, activities[0]['actor'])
     self.assertEqual(HTML_VIEWER_PUBLIC, activities[0]['object']['author'])
 
-  def test_html_to_activities_profile_wrong_id_dont_fill_in_owner(self):
+  def test_scraped_to_activities_profile_wrong_id_dont_fill_in_owner(self):
     profile = copy.deepcopy(HTML_PROFILE)
     user = profile['entry_data']['ProfilePage'][0]['graphql']['user']
     other_id = user['id'] + '999'
@@ -1557,7 +1592,7 @@ class InstagramTest(testutil.TestCase):
       'id': other_id,
     }
 
-    activities, _ = self.instagram.html_to_activities(
+    activities, _ = self.instagram.scraped_to_activities(
       HTML_HEADER + json_dumps(profile) + HTML_FOOTER)
     expected = {
       'id': tag_uri(other_id),
@@ -1566,23 +1601,46 @@ class InstagramTest(testutil.TestCase):
     self.assertEqual(expected, activities[0]['actor'])
     self.assertEqual(expected, activities[0]['object']['author'])
 
-  def test_html_to_activities_photo_no_fetch_extras(self):
-    activities, viewer = self.instagram.html_to_activities(
+  def test_scraped_to_activities_photo_no_fetch_extras(self):
+    activities, viewer = self.instagram.scraped_to_activities(
       HTML_PHOTO_COMPLETE, fetch_extras=False)
     self.assert_equals([HTML_PHOTO_ACTIVITY], activities)
     self.assertIsNone(viewer)
 
-  def test_html_to_activities_photo_fetch_extras(self):
+  def test_scraped_to_activities_photo_fetch_extras(self):
     self.expect_requests_get(
       instagram.HTML_LIKES_URL % 'ABC123', HTML_PHOTO_LIKES_RESPONSE, headers={})
     self.mox.ReplayAll()
 
-    activities, viewer = self.instagram.html_to_activities(
+    activities, viewer = self.instagram.scraped_to_activities(
       HTML_PHOTO_COMPLETE, fetch_extras=True)
     self.assert_equals([HTML_PHOTO_ACTIVITY_LIKES], activities)
     self.assertIsNone(viewer)
 
-  def test_html_to_activities_photo_edge_media_to_parent_comment(self):
+  def test_scraped_to_activity_photo_no_fetch_extras(self):
+    self.assert_equals(
+      (HTML_PHOTO_ACTIVITY, None),
+      self.instagram.scraped_to_activity(HTML_PHOTO_COMPLETE, fetch_extras=False))
+
+  def test_scraped_to_activity_photo_with_viewer(self):
+    page = copy.deepcopy(HTML_PHOTO_PAGE)
+    page['config'] = HTML_VIEWER_CONFIG
+    html = HTML_HEADER + json_dumps(page) + HTML_FOOTER
+
+    self.assert_equals((HTML_PHOTO_ACTIVITY, HTML_VIEWER),
+                       self.instagram.scraped_to_activity(html))
+
+  def test_scraped_to_activity_photo_fetch_extras(self):
+    self.expect_requests_get(
+      instagram.HTML_LIKES_URL % 'ABC123', HTML_PHOTO_LIKES_RESPONSE, headers={})
+    self.mox.ReplayAll()
+
+    activity, actor = self.instagram.scraped_to_activity(
+      HTML_PHOTO_COMPLETE, fetch_extras=True)
+    self.assert_equals(HTML_PHOTO_ACTIVITY_LIKES, activity)
+    self.assertIsNone(actor)
+
+  def test_scraped_to_activities_photo_edge_media_to_parent_comment(self):
     """https://github.com/snarfed/granary/issues/164"""
     self.expect_requests_get(
       instagram.HTML_LIKES_URL % 'ABC123', HTML_PHOTO_LIKES_RESPONSE, headers={})
@@ -1591,21 +1649,21 @@ class InstagramTest(testutil.TestCase):
     page = copy.deepcopy(HTML_PHOTO_PAGE)
     media = page['entry_data']['PostPage'][0]['graphql']['shortcode_media']
     media['edge_media_to_parent_comment'] = media.pop('edge_media_to_comment')
-    activities, _ = self.instagram.html_to_activities(
+    activities, _ = self.instagram.scraped_to_activities(
       HTML_HEADER + json_dumps(page) + HTML_FOOTER, fetch_extras=True)
     self.assert_equals([HTML_PHOTO_ACTIVITY_LIKES], activities)
 
-  def test_html_to_activities_video(self):
-    activities, viewer = self.instagram.html_to_activities(HTML_VIDEO_COMPLETE)
+  def test_scraped_to_activities_video(self):
+    activities, viewer = self.instagram.scraped_to_activities(HTML_VIDEO_COMPLETE)
     self.assert_equals([HTML_VIDEO_ACTIVITY_FULL], activities)
     self.assertIsNone(viewer)
 
-  def test_html_to_activities_multi_photo(self):
-    activities, viewer = self.instagram.html_to_activities(HTML_MULTI_PHOTO_COMPLETE)
+  def test_scraped_to_activities_multi_photo(self):
+    activities, viewer = self.instagram.scraped_to_activities(HTML_MULTI_PHOTO_COMPLETE)
     self.assert_equals([HTML_MULTI_PHOTO_ACTIVITY], activities)
     self.assertIsNone(viewer)
 
-  def test_html_to_activities_multi_photo_omit_auto_alt_text(self):
+  def test_scraped_to_activities_multi_photo_omit_auto_alt_text(self):
     multi = copy.deepcopy(HTML_MULTI_PHOTO_PAGE)
     multi['entry_data']['PostPage'][0]['graphql']['shortcode_media']\
         ['edge_sidecar_to_children']['edges'][1]['node']['accessibility_caption'] = \
@@ -1615,14 +1673,14 @@ class InstagramTest(testutil.TestCase):
     expected = copy.deepcopy(HTML_MULTI_PHOTO_ACTIVITY)
     del expected['object']['attachments'][1]['image'][0]['displayName']
 
-    activities, _ = self.instagram.html_to_activities(html)
+    activities, _ = self.instagram.scraped_to_activities(html)
     self.assert_equals([expected], activities)
 
-  def test_html_to_activities_missing_profile_picture_external_url(self):
+  def test_scraped_to_activities_missing_profile_picture_external_url(self):
     data = copy.deepcopy(HTML_FEED)
     data['config']['viewer']['profile_pic_url'] = None
     data['config']['viewer']['external_url'] = None
-    _, viewer = self.instagram.html_to_activities(
+    _, viewer = self.instagram.scraped_to_activities(
       HTML_HEADER + json_dumps(data) + HTML_FOOTER)
 
     expected = copy.deepcopy(HTML_VIEWER)
@@ -1630,11 +1688,11 @@ class InstagramTest(testutil.TestCase):
     del expected['image']
     self.assert_equals(expected, viewer)
 
-  def test_html_to_activities_missing_video_url(self):
+  def test_scraped_to_activities_missing_video_url(self):
     data = copy.deepcopy(HTML_FEED)
     del data['entry_data']['FeedPage'][0]['graphql']['user']\
             ['edge_web_feed_timeline']['edges'][1]['node']['video_url']
-    activities, _ = self.instagram.html_to_activities(
+    activities, _ = self.instagram.scraped_to_activities(
       HTML_HEADER + json_dumps(data) + HTML_FOOTER)
 
     expected = copy.deepcopy(HTML_ACTIVITIES_FULL)
@@ -1642,18 +1700,18 @@ class InstagramTest(testutil.TestCase):
     del expected[1]['object']['attachments'][0]['stream'][0]['url']
     self.assert_equals(expected, activities)
 
-  def test_html_to_activities_missing_header(self):
-    activities, viewer = self.instagram.html_to_activities(HTML_PHOTO_MISSING_HEADER)
+  def test_scraped_to_activities_missing_header(self):
+    activities, viewer = self.instagram.scraped_to_activities(HTML_PHOTO_MISSING_HEADER)
     self.assert_equals([], activities)
     self.assertIsNone(viewer)
 
-  def test_html_to_activities_missing_footer(self):
-    activities, viewer = self.instagram.html_to_activities(HTML_PHOTO_MISSING_FOOTER)
+  def test_scraped_to_activities_missing_footer(self):
+    activities, viewer = self.instagram.scraped_to_activities(HTML_PHOTO_MISSING_FOOTER)
     self.assert_equals([], activities)
     self.assertIsNone(viewer)
 
-  def test_html_to_activities_trims_nulls(self):
-    activities, viewer = self.instagram.html_to_activities(HTML_HEADER + json_dumps({
+  def test_scraped_to_activities_trims_nulls(self):
+    activities, viewer = self.instagram.scraped_to_activities(HTML_HEADER + json_dumps({
       'entry_data': {
         'FeedPage': [{
           'feed': {
@@ -1669,20 +1727,20 @@ class InstagramTest(testutil.TestCase):
     self.assert_equals([], activities)
     self.assertIsNone(viewer)
 
-  def test_html_to_activities_preload_fetch(self):
+  def test_scraped_to_activities_preload_fetch(self):
     """https://github.com/snarfed/granary/issues/140"""
     url = urllib.parse.urljoin(HTML_BASE_URL, HTML_PRELOAD_URL)
     self.expect_requests_get(url, HTML_PRELOAD_DATA, cookie='kuky')
     self.mox.ReplayAll()
 
     html = HTML_HEADER_PRELOAD + json_dumps(HTML_USELESS_FEED) + HTML_FOOTER
-    activities, viewer = self.instagram.html_to_activities(html, cookie='kuky')
+    activities, viewer = self.instagram.scraped_to_activities(html, cookie='kuky')
 
     self.assert_equals([HTML_PHOTO_ACTIVITY_FULL, HTML_VIDEO_ACTIVITY_FULL],
                        activities)
     self.assertIsNone(viewer)
 
-  def test_html_to_activities_preload_fetch_bad_json(self):
+  def test_scraped_to_activities_preload_fetch_bad_json(self):
     """https://console.cloud.google.com/errors/CP_w8ai-7JLfvAE"""
     url = urllib.parse.urljoin(HTML_BASE_URL, HTML_PRELOAD_URL)
     self.expect_requests_get(url, '{bad: ["json', cookie='kuky')
@@ -1690,17 +1748,28 @@ class InstagramTest(testutil.TestCase):
 
     html = HTML_HEADER_PRELOAD + json_dumps(HTML_USELESS_FEED) + HTML_FOOTER
     with self.assertRaises(requests.HTTPError) as cm:
-      self.instagram.html_to_activities(html, cookie='kuky')
+      self.instagram.scraped_to_activities(html, cookie='kuky')
 
     self.assertEqual(504, cm.exception.response.status_code)
 
+  def test_merge_scraped_reactions(self):
+    activity = copy.deepcopy(HTML_PHOTO_ACTIVITY)
+    got = self.instagram.merge_scraped_reactions(
+      json_dumps(HTML_PHOTO_LIKES_RESPONSE), activity)
+    self.assert_equals(LIKE_OBJS, got)
+    self.assert_equals(HTML_PHOTO_ACTIVITY_LIKES, activity)
+
+  def test_scraped_to_actor(self):
+    self.assert_equals(HTML_VIEWER_PUBLIC,
+                       self.instagram.scraped_to_actor(HTML_PROFILE_COMPLETE))
+    self.assertIsNone(self.instagram.scraped_to_actor(HTML_VIDEO_COMPLETE))
+
   def test_id_to_shortcode(self):
     for shortcode, id in (
         (None, None),
         (None, ''),
         ('BDJ7Nr5Nxpa', 1209758400153852506),
         ('BDJ7Nr5Nxpa', '1209758400153852506'),
-        ('BDJ7Nr5Nxpa', '1209758400153852506'),
         ('BDJ7Nr5Nxpa', '1209758400153852506_1103525'),
         ('BDJ7Nr5Nxpa', 'BDJ7Nr5Nxpa'),
         ('BDJ7N_5Nxpa', 'BDJ7N_5Nxpa'),
