@@ -403,6 +403,18 @@ class MastodonTest(testutil.TestCase):
     self.assert_equals([ACTIVITY, MEDIA_ACTIVITY],
                        self.mastodon.get_activities(fetch_mentions=True))
 
+  def test_get_activities_fetch_mentions_null_status(self):
+    """https://console.cloud.google.com/errors/CNvulo670obn4AE"""
+    self.expect_get(API_TIMELINE, params={}, response=[])
+
+    notif = copy.deepcopy(MENTION_NOTIFICATION)
+    notif['status'] = None
+    self.expect_get(API_NOTIFICATIONS, [notif], json={
+      'exclude_types': ['follow', 'favourite', 'reblog'],
+    })
+    self.mox.ReplayAll()
+    self.assert_equals([], self.mastodon.get_activities(fetch_mentions=True))
+
   def test_get_activities_activity_id(self):
     self.expect_get(API_STATUS % '123', STATUS)
     self.mox.ReplayAll()
@@ -420,17 +432,27 @@ class MastodonTest(testutil.TestCase):
     self.mox.ReplayAll()
     self.assert_equals([ACTIVITY], self.mastodon.get_activities(group_id=source.SELF))
 
-  def test_get_activities_search(self):
+  def test_get_activities_search_without_count(self):
     self.expect_get(API_SEARCH, params={
       'q': 'indieweb',
       'resolve': True,
       'offset': 0,
-      'limit': '',
     }, response={'statuses': [STATUS, MEDIA_STATUS]})
     self.mox.ReplayAll()
     self.assert_equals([ACTIVITY, MEDIA_ACTIVITY], self.mastodon.get_activities(
         group_id=source.SEARCH, search_query='indieweb'))
 
+  def test_get_activities_search_with_count(self):
+    self.expect_get(API_SEARCH, params={
+      'q': 'indieweb',
+      'resolve': True,
+      'offset': 0,
+      'limit': 123,
+    }, response={'statuses': [STATUS, MEDIA_STATUS]})
+    self.mox.ReplayAll()
+    self.assert_equals([ACTIVITY, MEDIA_ACTIVITY], self.mastodon.get_activities(
+        group_id=source.SEARCH, search_query='indieweb', count=123))
+
   def test_get_activities_search_no_query(self):
     with self.assertRaises(ValueError):
       self.mastodon.get_activities(group_id=source.SEARCH, search_query=None)
@@ -480,7 +502,7 @@ class MastodonTest(testutil.TestCase):
 
     self.mox.ReplayAll()
     cache = util.CacheDict()
-    for i in range(4):
+    for _ in range(4):
       self.mastodon.get_activities(fetch_replies=True, fetch_shares=True,
                                    fetch_likes=True, cache=cache)
 
@@ -567,6 +589,25 @@ class MastodonTest(testutil.TestCase):
     self.assertIsNone(result.error_plain)
     self.assertIsNone(result.error_html)
 
+  def test_create_bookmark(self):
+    self.expect_post(API_STATUSES, json={'status': 'foo ☕ bar'},
+                     response=STATUS)
+    self.mox.ReplayAll()
+
+    result = self.mastodon.create({
+      "objectType": "activity",
+      "verb": "post",
+      "content": 'foo ☕ bar',
+      "object": {
+        "objectType": "bookmark",
+        "targetUrl": "https://example.com/foo"
+      }
+    })
+
+    self.assert_equals(STATUS, result.content, result)
+    self.assertIsNone(result.error_plain)
+    self.assertIsNone(result.error_html)
+
   def test_create_reply(self):
     self.expect_post(API_STATUSES, json={
       'status': 'foo ☕ bar',
@@ -595,6 +636,19 @@ class MastodonTest(testutil.TestCase):
       'object': {'url': 'http://foo.com/@xyz/123'},
     }))
 
+  def test_base_object_no_url(self):
+    self.assert_equals({}, self.mastodon.base_object({
+      'object': {'foo': 'bar'},
+    }))
+
+  def test_base_object_bookmark(self):
+    self.assert_equals({}, self.mastodon.base_object({
+      'object': {
+        'objectType': 'bookmark',
+        'targetUrl': 'http://bar.com/baz',
+      },
+    }))
+
   def test_base_object_two_remote(self):
     bad = {'url': 'http://bad/456'}
     remote = {'url': STATUS_REMOTE['uri']}
@@ -656,7 +710,7 @@ class MastodonTest(testutil.TestCase):
 
     got = self.mastodon.create(LIKE).content
     self.assert_equals('like', got['type'])
-    self.assert_equals('http://foo.com/@snarfed/123', got['url'])
+    self.assert_equals('http://foo.com/@snarfed/123#favorited-by-23507', got['url'])
 
   def test_preview_favorite(self):
     preview = self.mastodon.preview_create(LIKE)
@@ -671,7 +725,7 @@ class MastodonTest(testutil.TestCase):
 
     got = self.mastodon.create(LIKE_REMOTE).content
     self.assert_equals('like', got['type'])
-    self.assert_equals(url, got['url'])
+    self.assert_equals(url + '#favorited-by-23507', got['url'])
 
   def test_preview_favorite_remote(self):
     url = STATUS_REMOTE['url']
@@ -724,13 +778,25 @@ class MastodonTest(testutil.TestCase):
     self.assertEqual('foo ☕ bar<br /><br /><video controls src="http://foo.com/video.mp4"><a href="http://foo.com/video.mp4">a fun video</a></video> &nbsp; <img src="http://foo.com/image.jpg" alt="" />',
                      preview.content)
 
+  def test_preview_create_override_truncate_text_length(self):
+    m = mastodon.Mastodon(INSTANCE, access_token='towkin',
+                          user_id=ACCOUNT['id'], truncate_text_length=8)
+    got = m.preview_create(OBJECT)
+    self.assertEqual('foo ☕…', got.content)
+
+    self.expect_post(API_STATUSES, json={'status': 'foo ☕…'}, response=STATUS)
+    self.mox.ReplayAll()
+
+    result = m.create(OBJECT)
+    self.assert_equals(STATUS, result.content, result)
+
   def test_create_with_media(self):
     self.expect_requests_get('http://foo.com/video.mp4', 'pic 2')
-    self.expect_post(API_MEDIA, {'id': 'a'}, files={'file': b'pic 2'})
+    self.expect_post(API_MEDIA, {'id': 'a'}, files={'file': b'pic 2'},
+                     data={'description': 'a fun video'})
 
     self.expect_requests_get('http://foo.com/image.jpg', 'pic 1')
-    self.expect_post(API_MEDIA, {'id': 'b'}, files={'file': b'pic 1'})
-
+    self.expect_post(API_MEDIA, {'id': 'b'}, files={'file': b'pic 1'}, data={})
     self.expect_post(API_STATUSES, json={
       'status': 'foo ☕ bar',
       'media_ids': ['a', 'b'],
@@ -764,10 +830,10 @@ class MastodonTest(testutil.TestCase):
 
     # test create
     self.expect_requests_get('http://my/video', 'vid')
-    self.expect_post(API_MEDIA, {'id': '0'}, files={'file': b'vid'})
+    self.expect_post(API_MEDIA, {'id': '0'}, files={'file': b'vid'}, data={})
     for i, url in enumerate(image_urls[:-1]):
       self.expect_requests_get('http://my/picture/%d' % i, 'pic')
-      self.expect_post(API_MEDIA, {'id': str(i + 1)}, files={'file': b'pic'})
+      self.expect_post(API_MEDIA, {'id': str(i + 1)}, files={'file': b'pic'}, data={})
 
     self.expect_post(API_STATUSES, json={
       'status': '',
