@@ -4,6 +4,7 @@
 Most of the tests are in testdata/. This is just a few things that are too small
 for full testdata tests.
 """
+import copy
 import re
 
 from oauth_dropins.webutil import testutil
@@ -286,7 +287,7 @@ bar<br />
   def test_render_content_omits_tags_without_urls(self):
     self.assert_equals("""\
 foo
-<a class="tag" href="http://baj"></a>
+<a class="tag" aria-hidden="true" href="http://baj"></a>
 <a class="tag" href="http://baz">baz</a>
 """, microformats2.render_content({
       'content': 'foo',
@@ -409,8 +410,6 @@ Shared <a href="#">a post</a> by foo
         'displayName': 'd & e'}],
     }
 
-    # TODO: test that img alt gets displayName 'd & e' once mf2py handles that.
-    # https://github.com/tommorris/mf2py/issues/83
     self.assert_multiline_equals("""\
 <article class="h-entry">
 <span class="p-uid"></span>
@@ -428,13 +427,12 @@ Shared <a href="#">a post</a> by foo
     self.assert_multiline_equals("""\
 <p>
 <img class="u-photo" src="att-img" alt="d &amp; e" />
-<span class="name">d & e</span>
 </p>""", content, ignore_blanks=True)
 
   def test_mention_and_hashtag(self):
     self.assert_equals("""
 <a class="p-category" href="http://c">c</a>
-<a class="u-mention" href="http://m"></a>""",
+<a class="u-mention" aria-hidden="true" href="http://m"></a>""",
                        microformats2.render_content({
         'tags': [{'objectType': 'mention', 'url': 'http://m', 'displayName': 'm'},
                  {'objectType': 'hashtag', 'url': 'http://c', 'displayName': 'c'}],
@@ -443,8 +441,8 @@ Shared <a href="#">a post</a> by foo
   def test_tag_multiple_urls(self):
     expected_urls = ['http://1', 'https://2']
     expected_html = """
-<a class="tag" href="http://1"></a>
-<a class="tag" href="https://2"></a>
+<a class="tag" aria-hidden="true" href="http://1"></a>
+<a class="tag" aria-hidden="true" href="https://2"></a>
 """
     for tag in ({'url': 'http://1',
                   'urls': [{'value': 'http://1'}, {'value': 'https://2'}]},
@@ -482,7 +480,7 @@ Shared <a href="#">a post</a> by foo
             'url': ['http://p'],
           },
         }],
-        'content': [{'html': '\n<a class="tag" href="http://x"></a>'}],
+        'content': [{'html': '\n<a class="tag" aria-hidden="true" href="http://x"></a>'}],
       },
     }, microformats2.object_to_json(obj))
 
@@ -576,6 +574,20 @@ Shared <a href="#">a post</a> by foo
     """
     self.assert_equals({}, microformats2.object_to_json('foo bar'))
 
+  def test_object_to_json_content_null(self):
+    """https://console.cloud.google.com/errors/CNzPuIvvlbbd3QE"""
+    self.assert_equals({
+      'type': ['h-entry'],
+      'properties': {
+        'uid': ['tag:fake.com:10157']
+      },
+    }, microformats2.object_to_json({
+      'objectType': 'activity',
+      'verb': 'react',
+      'id': 'tag:fake.com:10157',
+      'content': None,
+    }))
+
   def test_get_string_urls(self):
     for expected, objs in (
         ([], []),
@@ -925,3 +937,28 @@ Shared <a href="#">a post</a> by   <span class="h-card">
         ('5 MB', 5000000),
     ):
       self.assertEqual(expected, microformats2.size_to_bytes(input), input)
+
+  def test_prefix_image_urls(self):
+    IMG = lambda: copy.deepcopy({'image': {'url': 'xyz'}})
+    PRE = lambda: copy.deepcopy({'image': {'url': 'abcxyz'}})
+
+    activity = {
+      'actor': IMG(),
+      'object': {
+        'author': IMG(),
+        'replies': {'items': [IMG(), IMG(), PRE()]},
+        'attachments': [IMG(), {'author': IMG()}],
+        'tags': [IMG(), {'actor': IMG()}],
+      },
+    }
+    microformats2.prefix_image_urls(activity, 'abc')
+
+    self.assert_equals({
+      'actor': PRE(),
+      'object': {
+        'author': PRE(),
+        'replies': {'items': [PRE(), PRE(), PRE()]},
+        'attachments': [PRE(), {'author': PRE()}],
+        'tags': [PRE(), {'actor': PRE()}],
+      },
+    }, activity)
