@@ -130,6 +130,11 @@ class FakeSource(Source):
   EMBED_POST = 'foo %(url)s bar'
 
 
+class FakeHTMLSource(Source):
+  DOMAIN = 'fake.com'
+  EMBED_POST = '%(content)s'
+
+
 class SourceTest(testutil.TestCase):
 
   def setUp(self):
@@ -239,6 +244,9 @@ class SourceTest(testutil.TestCase):
            'tags': [{'url': 'http://bad]'}]},
           [])
 
+    # bookmarks should include targetUrl
+    check({'targetUrl': 'http://or.ig/'}, ['http://or.ig/'])
+
   def test_original_post_discovery_follow_redirects(self):
     self.expect_requests_head('http://other/link',
                               redirected_url='http://other/link/redirected'
@@ -261,6 +269,39 @@ class SourceTest(testutil.TestCase):
     check(obj, ['http://or.ig/post/redirected', 'http://other/link/redirected'],
           include_redirect_sources=False)
 
+  def test_original_post_discovery_excludes(self):
+    """Should exclude reserved hosts, non-http(s) URLs, and missing domains."""
+    obj = {
+      'content': 'foo',
+      'upstreamDuplicates': [
+        'http://sho.rt/post',
+        # local
+        'http://localhost',
+        'http://other/link',
+        'http://y.local/path'
+        # reserved
+        'https://x.test/',
+        # not http
+        'file://foo.com/Ryan/.npmrc'
+        'git@github.com:snarfed/granary.git',
+        '/home/ryan/foo',
+        'mailto:x@y.z',
+        # missing domain
+        'http:///foo/bar',
+        'file:///Users/Ryan/.npmrc',
+      ],
+    }
+    self.check_original_post_discovery(
+      obj, ['http://sho.rt/post'], include_reserved_hosts=False)
+
+  def test_original_post_discovery_exclude__hosts(self):
+    obj = {
+      'content': 'http://other/link https://x.test/ http://y.local/path',
+      'upstreamDuplicates': ['http://localhost', 'http://sho.rt/post'],
+    }
+    self.check_original_post_discovery(
+      obj, ['http://sho.rt/post'], include_reserved_hosts=False)
+
   def test_get_like(self):
     self.source.get_activities(user_id='author', activity_id='activity',
                                fetch_likes=True).AndReturn([ACTIVITY])
@@ -483,6 +524,11 @@ Watching  \t waves
       self.assertFalse(self.source.activity_changed(before, after, log=True),
                                                     '%s\n%s' % (before, after))
 
+    fb_comment_edited_inReplyTo = copy.deepcopy(fb_comment_edited)
+    fb_comment_edited_inReplyTo['inReplyTo'].append({
+      'id': 'tag:fake.com:000000000000000',
+      'url': 'https://www.facebook.com/000000000000000',
+    })
     fb_comment_edited['content'] = 'new content'
     gp_like_edited['to'] = [{'objectType':'group', 'alias':'@private'}]
 
@@ -491,11 +537,27 @@ Watching  \t waves
     fb_rsvp = RSVP_YES
 
     for before, after in ((fb_comment, fb_comment_edited),
+                          (fb_comment, fb_comment_edited_inReplyTo),
                           (gp_like, gp_like_edited),
                           (fb_invite, fb_rsvp)):
       self.assertTrue(self.source.activity_changed(before, after, log=True),
                                                    '%s\n%s' % (before, after))
 
+  def test_append_in_reply_to(self):
+    fb_comment_before = copy.deepcopy(COMMENT)
+    fb_comment_after_same = copy.deepcopy(fb_comment_before)
+    self.source.append_in_reply_to(fb_comment_before,fb_comment_after_same)
+    self.assertEqual(COMMENT,fb_comment_before)
+    self.assertEqual(COMMENT,fb_comment_after_same)
+
+    fb_comment_after_diff = copy.deepcopy(fb_comment_before)
+    fb_comment_after_targ = copy.deepcopy(fb_comment_before)
+    fb_comment_after_diff['inReplyTo'] = ['new']
+    fb_comment_after_targ['inReplyTo'] = fb_comment_after_diff.get('inReplyTo')+fb_comment_before.get('inReplyTo')
+    self.source.append_in_reply_to(fb_comment_before,fb_comment_after_diff)
+    self.assertEqual(fb_comment_after_targ,fb_comment_after_diff)
+
+
   def test_sources_global(self):
     self.assertEqual(facebook.Facebook, source.sources['facebook'])
     self.assertEqual(instagram.Instagram, source.sources['instagram'])
@@ -556,9 +618,9 @@ Watching  \t waves
     self.assert_equals('foo http://%3Ca%3Eb bar',
                        self.source.embed_post({'url': 'http://<a>b'}))
 
-  def test_embed_post_checks_content_for_html(self):
-    with self.assertRaises(ValueError):
-      self.source.embed_post({'content': '<xyz>'})
+  def test_embed_post_escapes_html(self):
+    self.assert_equals('x &gt; 2 &amp;&amp; x &lt; 7',
+      FakeHTMLSource().embed_post({'content': 'x > 2 && x < 7' }))
 
   def test_truncate(self):
     """A bunch of tests to exercise the text shortening algorithm."""
