@@ -3,8 +3,6 @@
 
 Uses the v1.1 REST API: https://developer.twitter.com/en/docs/api-reference-index
 
-TODO: collections for twitter accounts; use as activity target?
-
 The Audience Targeting 'to' field is set to @public or @private based on whether
 the tweet author's 'protected' field is true or false.
 https://dev.twitter.com/docs/platform-objects/users
@@ -17,11 +15,13 @@ import logging
 import mimetypes
 import re
 import socket
+import time
 import urllib.parse, urllib.request
 
 from oauth_dropins import twitter_auth
 from oauth_dropins.webutil import util
 from oauth_dropins.webutil.util import json_dumps, json_loads
+from requests import RequestException
 
 from . import source
 
@@ -36,6 +36,7 @@ API_DELETE_TWEET = 'statuses/destroy.json'
 API_DELETE_FAVORITE = 'favorites/destroy.json'
 API_FAVORITES = 'favorites/list.json?screen_name=%s'
 API_LIST_TIMELINE = 'lists/statuses.json?count=%(count)d&slug=%(slug)s&owner_screen_name=%(owner_screen_name)s' + API_TWEET_PARAMS
+API_LIST_ID_TIMELINE = 'lists/statuses.json?count=%(count)d&list_id=%(list_id)s' + API_TWEET_PARAMS
 API_LOOKUP = 'statuses/lookup.json?id=%s' + API_TWEET_PARAMS
 API_POST_FAVORITE = 'favorites/create.json'
 API_POST_MEDIA = 'statuses/update_with_media.json'
@@ -46,10 +47,12 @@ API_SEARCH = 'search/tweets.json?q=%(q)s&result_type=recent&count=%(count)d' + A
 API_STATUS = 'statuses/show.json?id=%s' + API_TWEET_PARAMS
 API_TIMELINE = 'statuses/home_timeline.json?count=%d' + API_TWEET_PARAMS
 API_UPLOAD_MEDIA = 'https://upload.twitter.com/1.1/media/upload.json'
+API_MEDIA_STATUS_MAX_DELAY_SECS = 30
+API_MEDIA_STATUS_MAX_POLLS = 20
 API_MEDIA_METADATA = 'https://upload.twitter.com/1.1/media/metadata/create.json'
 API_USER = 'users/show.json?screen_name=%s'
 API_USER_TIMELINE = 'statuses/user_timeline.json?count=%(count)d&screen_name=%(screen_name)s' + API_TWEET_PARAMS
-HTML_FAVORITES = 'https://twitter.com/i/activity/favorited_popup?id=%s'
+SCRAPE_LIKES_URL = 'https://api.twitter.com/2/timeline/liked_by.json?tweet_mode=extended&include_user_entities=true&tweet_id=%s&count=80'
 
 TWEET_URL_RE = re.compile(r'https://twitter\.com/[^/?]+/status(es)?/[^/?]+$')
 HTTP_RATE_LIMIT_CODES = (429, 503)
@@ -100,13 +103,18 @@ MAX_ALT_LENGTH = 420
 # username requirements and limits:
 # https://support.twitter.com/articles/101299#error
 # http://stackoverflow.com/a/13396934/186123
-MENTION_RE = re.compile(r'(^|[^\w@/\!?=&])@(\w{1,15})\b', re.UNICODE)
+USERNAME = r'\w{1,15}'
+USERNAME_RE = re.compile(USERNAME + '$')
+MENTION_RE = re.compile(r'(^|[^\w@/\!?=&])@(' + USERNAME + r')\b', re.UNICODE)
 
 # hashtag requirements and limits:
 # https://support.twitter.com/articles/370610
 # http://stackoverflow.com/questions/8451846
 HASHTAG_RE = re.compile(r'(^|\s)[#＃](\w+)\b', re.UNICODE)
 
+# alias allows unit tests to mock this function
+sleep_fn = time.sleep
+
 
 class OffsetTzinfo(datetime.tzinfo):
   """A simple, DST-unaware tzinfo from given utc offset in seconds.
@@ -134,6 +142,7 @@ class Twitter(source.Source):
   NAME = 'Twitter'
   FRONT_PAGE_TEMPLATE = 'templates/twitter_index.html'
   POST_ID_RE = re.compile('^[0-9]+$')
+  OPTIMIZED_COMMENTS = True
 
   # HTML snippet for embedding a tweet.
   # https://dev.twitter.com/docs/embedded-tweets
@@ -155,7 +164,8 @@ class Twitter(source.Source):
   # TRUNCATE_TEXT_LENGTH = None
   # TRUNCATE_URL_LENGTH = None
 
-  def __init__(self, access_token_key, access_token_secret, username=None):
+  def __init__(self, access_token_key, access_token_secret, username=None,
+               scrape_headers=None):
     """Constructor.
 
     Twitter now requires authentication in v1.1 of their API. You can get an
@@ -165,10 +175,13 @@ class Twitter(source.Source):
       access_token_key: string, OAuth access token key
       access_token_secret: string, OAuth access token secret
       username: string, optional, the current user. Used in e.g. preview/create.
+      scrape_headers: dict, optional, with string HTTP header keys and values to
+        use when scraping likes
     """
     self.access_token_key = access_token_key
     self.access_token_secret = access_token_secret
     self.username = username
+    self.scrape_headers = scrape_headers
 
   def get_actor(self, screen_name=None):
     """Returns a user as a JSON ActivityStreams actor dict.
@@ -176,10 +189,7 @@ class Twitter(source.Source):
     Args:
       screen_name: string username. Defaults to the current user.
     """
-    if screen_name is None:
-      url = API_CURRENT_USER
-    else:
-      url = API_USER % screen_name
+    url = API_CURRENT_USER if screen_name is None else API_USER % screen_name
     return self.user_to_actor(self.urlopen(url))
 
   def get_activities_response(self, user_id=None, group_id=None, app_id=None,
@@ -205,7 +215,7 @@ class Twitter(source.Source):
     "I've confirmed with our team that we're not explicitly supporting this
     family of features."
 
-    Likes (ie favorites) are scraped from twitter.com HTML, since Twitter's REST
+    Likes (nee favorites) are scraped from twitter.com, since Twitter's REST
     API doesn't offer a way to fetch them. You can also get them from the
     Streaming API, though, and convert them with streaming_event_to_object().
     https://dev.twitter.com/docs/streaming-apis/messages#Events_event
@@ -238,15 +248,26 @@ class Twitter(source.Source):
     * it's not a reply, OR
     * it's a reply, but not to the current user, AND
       * the tweet it's replying to doesn't @-mention the current user
+
+    Raises:
+      NotImplementedError: if fetch_likes is True but scrape_headers was not
+        provided to the constructor.
     """
+    if fetch_likes and not self.scrape_headers:
+        raise NotImplementedError('fetch_likes requires scrape_headers')
+
     if group_id is None:
       group_id = source.FRIENDS
 
-    if user_id and user_id.startswith('@'):
-      user_id = user_id[1:]
+    if user_id:
+      if user_id.startswith('@'):
+        user_id = user_id[1:]
+      if not USERNAME_RE.match(user_id):
+        raise ValueError(f'Invalid Twitter username: {user_id}')
 
     # nested function for lazily fetching the user object if we need it
     user = []
+
     def _user():
       if not user:
         user.append(self.urlopen(API_USER % user_id if user_id else API_CURRENT_USER))
@@ -283,13 +304,21 @@ class Twitter(source.Source):
       elif group_id in (source.FRIENDS, source.ALL):
         url = API_TIMELINE % (count)
       else:
-        if not user_id:
-          user_id = _user().get('screen_name')
-        url = API_LIST_TIMELINE % {
-          'count': count,
-          'slug': group_id,
-          'owner_screen_name': user_id,
-        }
+        if util.is_int(group_id):
+          # it's a list id
+          url = API_LIST_ID_TIMELINE % {
+            'count': count,
+            'list_id': group_id,
+          }
+        else:
+          # it's a list slug
+          if not user_id:
+            user_id = _user().get('screen_name')
+          url = API_LIST_TIMELINE % {
+            'count': count,
+            'slug': urllib.parse.quote(group_id),
+            'owner_screen_name': user_id,
+          }
 
       headers = {'If-None-Match': etag} if etag else {}
       total_count = None
@@ -379,14 +408,16 @@ class Twitter(source.Source):
         id = tweet['id_str']
         count = tweet.get('favorite_count')
         if self.is_public(activity) and count and count != cached.get('ATF ' + id):
-          url = HTML_FAVORITES % id
           try:
-            resp = util.urlopen(url).read()
-            html = source.load_json(resp, url).get('htmlUsers', '')
-          except urllib.error.URLError as e:
+            resp = util.requests_get(SCRAPE_LIKES_URL % id,
+                                     headers=self.scrape_headers)
+            resp.raise_for_status()
+          except RequestException as e:
             util.interpret_http_exception(e)  # just log it
             continue
-          likes = self.favorites_html_to_likes(tweet, html)
+
+          likes = [self._make_like(tweet, author) for author in
+                   resp.json().get('globalObjects', {}).get('users', {}).values()]
           activity['object'].setdefault('tags', []).extend(likes)
           cache_updates['ATF ' + id] = count
 
@@ -456,7 +487,7 @@ class Twitter(source.Source):
       activity['object']['replies'] = {
         'items': items,
         'totalItems': len(items),
-        }
+      }
 
   def fetch_mentions(self, username, tweets, min_id=None):
     """Fetches a user's @-mentions and returns them as ActivityStreams.
@@ -742,15 +773,6 @@ class Twitter(source.Source):
     if base_id and not base_url:
       base_url = 'https://twitter.com/-/statuses/' + base_id
 
-    if is_reply and not base_url:
-      return source.creation_result(
-        abort=True,
-        error_plain='Could not find a tweet to reply to.',
-        error_html='Could not find a tweet to <a href="http://indiewebcamp.com/reply">reply to</a>. '
-        'Check that your post has an <a href="http://indiewebcamp.com/comment">in-reply-to</a> '
-        'link to a Twitter URL or to an original post that publishes a '
-        '<a href="http://indiewebcamp.com/rel-syndication">rel-syndication</a> link to Twitter.')
-
     # truncate and ellipsize content if it's over the character
     # count. URLs will be t.co-wrapped, so include that when counting.
     content = self.truncate(content, obj.get('url'), include_link, type=type,
@@ -803,11 +825,12 @@ class Twitter(source.Source):
         resp = self.urlopen(API_POST_RETWEET % base_id, data=data)
         resp['type'] = 'repost'
 
-    elif type in ('note', 'article') or is_reply or is_rsvp:  # a tweet
+    elif (type in ('note', 'article') or is_reply or is_rsvp or
+          (type == 'activity' and verb == 'post')):  # probably a bookmark
       content = str(content).encode('utf-8')
       data = [('status', content)]
 
-      if is_reply:
+      if is_reply and base_url:
         preview_description += """\
 <span class="verb">@-reply</span> to <a href="%s">this tweet</a>:
 %s""" % (base_url, self.embed_post(base_obj))
@@ -854,9 +877,8 @@ class Twitter(source.Source):
       if preview:
         return source.creation_result(content=preview_content,
                                       description=preview_description)
-      else:
-        resp = self.urlopen(API_POST_TWEET, data=urllib.parse.urlencode(sorted(data)))
-        resp['type'] = 'comment' if is_reply else 'post'
+      resp = self.urlopen(API_POST_TWEET, data=urllib.parse.urlencode(sorted(data)))
+      resp['type'] = 'comment' if is_reply and base_url else 'post'
 
     else:
       return source.creation_result(
@@ -915,9 +937,10 @@ class Twitter(source.Source):
         alt = util.ellipsize(alt, chars=MAX_ALT_LENGTH)
         headers = twitter_auth.auth_header(
           API_MEDIA_METADATA, self.access_token_key, self.access_token_secret, 'POST')
-        resp = util.requests_post(API_MEDIA_METADATA,
-                                  json={'media_id': media_id,'alt_text': {'text': alt}},
-                                  headers=headers)
+        resp = util.requests_post(
+          API_MEDIA_METADATA,
+          json={'media_id': media_id, 'alt_text': {'text': alt}},
+          headers=headers)
         resp.raise_for_status()
         logging.info('Got: %s', resp.text)
 
@@ -932,8 +955,7 @@ class Twitter(source.Source):
     * command=APPEND for each 5MB block, up to 15MB total
     * command=FINALIZE
 
-    https://dev.twitter.com/rest/reference/post/media/upload-chunked
-    https://dev.twitter.com/rest/public/uploading-media#chunkedupload
+    https://developer.twitter.com/en/docs/media/upload-media/uploading-media/chunked-media-upload
 
     Args:
       url: string URL of images
@@ -951,6 +973,8 @@ class Twitter(source.Source):
     media_id = self.urlopen(API_UPLOAD_MEDIA, data=urllib.parse.urlencode({
       'command': 'INIT',
       'media_type': 'video/mp4',
+      # https://twittercommunity.com/t/large-file-can-not-be-finalized-synchronously/82929/3
+      'media_category': 'tweet_video',
       # _check_media checked that Content-Length is set
       'total_bytes': video_resp.headers['Content-Length'],
     }))['media_id_string']
@@ -976,12 +1000,36 @@ class Twitter(source.Source):
       i += 1
 
     # FINALIZE
-    self.urlopen(API_UPLOAD_MEDIA, data=urllib.parse.urlencode({
+    resp = self.urlopen(API_UPLOAD_MEDIA, data=urllib.parse.urlencode({
       'command': 'FINALIZE',
       'media_id': media_id,
     }))
 
-    return media_id
+    total_delay = 0
+    for _ in range(API_MEDIA_STATUS_MAX_POLLS):
+      info = resp.get('processing_info', {})
+      state = info.get('state')
+      if not state or state == 'succeeded':
+        return media_id
+      elif state == 'failed':
+        # TODO test
+        return source.creation_result(abort=True, error_plain=str(info.get('error')))
+
+      # STATUS
+      delay = min(info.get('check_after_secs', 0),
+                  API_MEDIA_STATUS_MAX_DELAY_SECS)
+      total_delay += delay
+      logging.info(f'video still processing, waiting {delay}s to check status')
+      sleep_fn(delay)
+
+      params = urllib.parse.urlencode({
+        'command': 'STATUS',
+        'media_id': media_id,
+      })
+      resp = self.urlopen(f'{API_UPLOAD_MEDIA}?{params}')
+
+    msg = f'Twitter still processing uploaded video after {total_delay}s'
+    return source.creation_result(abort=True, error_plain=msg)
 
   @staticmethod
   def _check_media(url, resp, types, label, max_size):
@@ -1018,7 +1066,6 @@ class Twitter(source.Source):
         length // MB, max_size // MB, util.pretty_link(url))
       return source.creation_result(abort=True, error_plain=msg, error_html=msg)
 
-
   def delete(self, id):
     """Deletes a tweet. The authenticated user must have authored it.
 
@@ -1063,7 +1110,7 @@ class Twitter(source.Source):
         except http.client.HTTPException as e:
           if not str(e).startswith('Deadline exceeded'):
             raise
-        except socket.timeout as e:
+        except socket.timeout:
           pass
         except urllib.error.HTTPError as e:
           code, body = util.interpret_http_exception(e)
@@ -1123,7 +1170,7 @@ class Twitter(source.Source):
       'url': obj.get('url'),
       'actor': obj.get('author'),
       'object': obj,
-      }
+    }
 
     retweeted = tweet.get('retweeted_status')
     if retweeted:
@@ -1177,10 +1224,16 @@ class Twitter(source.Source):
     }
 
     retweeted = tweet.get('retweeted_status')
-    base_tweet = retweeted if retweeted else tweet
+    base_tweet = retweeted or tweet
     entities = self._get_entities(base_tweet)
-    text = util.WideUnicode(base_tweet.get('text') or '')
 
+    # text content
+    text = util.WideUnicode(
+      base_tweet.get('full_text') or base_tweet.get('text') or '')
+    text_start, text_end = (base_tweet.get('display_text_range')
+                            or (0, len(text)))
+
+    # author
     user = tweet.get('user')
     if user:
       obj['author'] = self.user_to_actor(user)
@@ -1206,7 +1259,10 @@ class Twitter(source.Source):
       }
       obj['attachments'] = [{
           'objectType': types.get(m.get('type')),
-          'image': {'url': m.get('media_url_https') or m.get('media_url')},
+          'image': {
+            'url': m.get('media_url_https') or m.get('media_url'),
+            'displayName': m.get('ext_alt_text'),
+          },
           'stream': {'url': self._video_url(m)},
           'displayName': m.get('ext_alt_text'),
       } for m in media]
@@ -1223,15 +1279,19 @@ class Twitter(source.Source):
     if quoted:
       quoted_obj = self.tweet_to_object(quoted)
       obj.setdefault('attachments', []).append(quoted_obj)
-      quoted_url = quoted_obj.get('url')
+      quoted_url = (quoted_obj.get('url') or
+                    tweet.get('quoted_status_permalink', {}).get('expanded'))
 
-      # remove quoted tweet URL from text
+      # remove quoted tweet URL from text, tags
       url_entities = entities.get('urls', [])
       for i, entity in enumerate(url_entities):
         indices = entity.get('indices')
         if indices and entity.get('expanded_url') == quoted_url:
-          text = text[:indices[0]] + text[indices[1]:]
+          start, end = indices
+          text = text[:start] + text[end:]
           del url_entities[i]
+          if start >= text_start and end <= text_end:
+            text_end -= (end - start)
 
     # tags
     obj['tags'] = [
@@ -1277,22 +1337,12 @@ class Twitter(source.Source):
       rt_prefix = 'RT <a href="https://twitter.com/%s">@%s</a>: ' % (
         (retweeted.get('user', {}).get('screen_name'),) * 2)
 
-    # text content. linkify entities. convert start/end indices to start/length,
-    # and replace t.co URLs with real "display" URLs.
-    text_start, text_end = (0, len(text))
+    # person mentions
+    obj['to'].extend(tag for tag in obj['tags']
+                     if tag.get('objectType') in ('person', 'mention')
+                     and tag.get('indices')[1] <= text_start)
 
-    full_text = base_tweet.get('full_text')
-    if full_text:
-      text = util.WideUnicode(full_text)
-      text_start, text_end = (tweet['display_text_range']
-                              if tweet.get('display_text_range')
-                              else (0, len(text)))
-      obj['to'].extend(tag for tag in obj['tags']
-                       if tag.get('objectType') in ('person', 'mention')
-                       and tag.get('indices')[1] <= text_start)
-
-    # convert start/end indices to start/length, and replace t.co URLs with
-    # real "display" URLs.
+    # replace entities with display URLs, convert start/end indices to start/length
     content = util.WideUnicode(rt_prefix + text[text_start:text_end])
     offset = len(rt_prefix) - text_start
     for t in obj['tags']:
@@ -1321,7 +1371,7 @@ class Twitter(source.Source):
       obj['location'] = {
         'displayName': place.get('full_name'),
         'id': self.tag_uri(place.get('id')),
-        }
+      }
 
       # place['url'] is a JSON API url, not useful for end users. get the
       # lat/lon from geo instead.
@@ -1337,9 +1387,9 @@ class Twitter(source.Source):
     reply_to_id = tweet.get('in_reply_to_status_id')
     if reply_to_id and reply_to_screenname:
       obj['inReplyTo'] = [{
-          'id': self.tag_uri(reply_to_id),
-          'url': self.status_url(reply_to_screenname, reply_to_id),
-          }]
+        'id': self.tag_uri(reply_to_id),
+        'url': self.status_url(reply_to_screenname, reply_to_id),
+      }]
 
     return self.postprocess_object(obj)
 
@@ -1358,15 +1408,17 @@ class Twitter(source.Source):
     """
     entities = collections.defaultdict(list)
 
-    # maps kind to set of id_str, url, and text values we've seen, for de-duping
+    # maps kind to set of id_str, url, and text values we've seen, with indices,
+    # for de-duping
     seen_ids = collections.defaultdict(set)
 
     for field in 'extended_entities', 'entities':  # prefer extended_entities!
       # kind is media, urls, hashtags, user_mentions, symbols, etc
       for kind, values in tweet.get(field, {}).items():
         for v in values:
-          id = v.get('id_str') or v.get('id') or v.get('url') or v.get('text')
-          if id:
+          id = (v.get('id_str') or v.get('id') or v.get('url') or v.get('text'),
+                tuple(v.get('indices') or []))
+          if id[0] or id[1]:
             if id in seen_ids[kind]:
               continue
             seen_ids[kind].add(id)
@@ -1447,7 +1499,7 @@ class Twitter(source.Source):
       'location': {'displayName': user.get('location')},
       'username': username,
       'description': user.get('description'),
-      })
+    })
 
   def retweet_to_object(self, retweet):
     """Converts a retweet to a share activity object.
@@ -1467,7 +1519,7 @@ class Twitter(source.Source):
         'objectType': 'activity',
         'verb': 'share',
         'object': {'url': self.tweet_url(orig)},
-        })
+    })
     if 'tags' in share:
       # the existing tags apply to the original tweet's text, which we replaced
       del share['tags']
@@ -1493,43 +1545,6 @@ class Twitter(source.Source):
       obj['published'] = self.rfc2822_to_iso8601(event.get('created_at'))
       return obj
 
-  def favorites_html_to_likes(self, tweet, html):
-    """Converts the HTML from a favorited_popup request to like objects.
-
-    e.g. https://twitter.com/i/activity/favorited_popup?id=434753879708672001
-
-    Args:
-      html: string
-
-    Returns:
-      list of ActivityStreams like object dicts
-    """
-    soup = util.parse_html(html)
-    likes = []
-
-    for user in soup.find_all(class_='js-user-profile-link'):
-      username = user.find(class_='username')
-      if not username:
-        continue
-      username = str(username.get_text(''))
-      if username.startswith('@'):
-        username = username[1:]
-
-      img = user.find(class_='js-action-profile-avatar') or {}
-      author = {
-        'id_str': img.get('data-user-id'),
-        'screen_name': username,
-        'profile_image_url': img.get('src'),
-        }
-
-      fullname = user.find(class_='fullname')
-      if fullname:
-        author['name'] = fullname.get_text(' ', strip=True)
-
-      likes.append(self._make_like(tweet, author))
-
-    return likes
-
   def _make_like(self, tweet, liker):
     """Generates and returns a ActivityStreams like object.
 
@@ -1557,7 +1572,7 @@ class Twitter(source.Source):
         'verb': 'like',
         'object': {'url': obj_url},
         'author': self.user_to_actor(liker),
-        })
+    })
 
   @staticmethod
   def rfc2822_to_iso8601(time_str):
@@ -1574,9 +1589,9 @@ class Twitter(source.Source):
 
     without_timezone = re.sub(' [+-][0-9]{4} ', ' ', time_str)
     timezone = re.search('[+-][0-9]{4}', time_str).group(0)
-    ## convert offset to seconds
+    # convert offset to seconds
     offset = 3600 * int(timezone[1:3]) + 60 * int(timezone[3:])
-    ## negative offset
+    # negative offset
     if timezone[0] == '-':
       offset = -offset
 
