@@ -1,18 +1,25 @@
 """Unit tests for logs.py. Woefully incomplete."""
 import datetime
-import unittest
+import time
 
+from flask import Flask
 from google.cloud import ndb
-
 from mox3 import mox
 
-from .. import logs
+from .. import appengine_config, logs
+
+with appengine_config.ndb_client.context():
+  KEY = ndb.Key('Foo', 123)
+  KEY_STR = KEY.urlsafe().decode()
 
-# KEY = ndb.Key('Foo', 123)
-# KEY_STR = KEY.urlsafe().decode()
 
-@unittest.skip
 class LogsTest(mox.MoxTestBase):
+  def setUp(self):
+    super().setUp()
+    self.app = Flask('test_logs')
+    self.app.add_url_rule('/log', view_func=logs.log)
+    self.app.config['TESTING'] = True
+    self.client = self.app.test_client()
 
   def test_url(self):
     self.assertEqual('log?start_time=172800&key=%s' % KEY_STR,
@@ -20,9 +27,9 @@ class LogsTest(mox.MoxTestBase):
 
   def test_maybe_link(self):
     when = datetime.datetime(1970, 1, 3)
-    expected = r'<time class="foo" datetime="1970-01-03T00:00:00" title="Sat Jan  3 00:00:00 1970">\d+ years ago</time>'
+    expected = r'<time class="foo" datetime="1970-01-03T00:00:00\+00:00" title="Sat Jan  3 00:00:00 1970 UTC">\d+ years ago</time>'
     actual = logs.maybe_link(when, KEY, time_class='foo')
-    self.assertRegexpMatches(actual, expected)
+    self.assertRegex(actual, expected)
 
     self.mox.StubOutWithMock(logs, 'MAX_LOG_AGE')
     logs.MAX_LOG_AGE = datetime.timedelta(days=99999)
@@ -35,3 +42,12 @@ class LogsTest(mox.MoxTestBase):
     when = datetime.datetime.now() + datetime.timedelta(minutes=1)
     got = logs.maybe_link(when, KEY)
     self.assertFalse(got.startswith('<a'), got)
+
+  def test_utcfromtimestamp_overflow(self):
+    too_big = 999999999999999999999
+    with self.assertRaises(OverflowError):
+      time.gmtime(too_big)
+
+    resp = self.client.get(f'/log?key=abc&start_time={too_big}')
+    self.assertEqual(400, resp.status_code)
+    self.assertIn('start_time too big', resp.get_data(as_text=True))
