@@ -11,10 +11,14 @@ import io
 from urllib.error import HTTPError, URLError
 import urllib.parse, urllib.request
 
+import prawcore.exceptions
 import requests
+import tumblpy
+import tweepy
 import urllib3
 import webapp2
 from webob import exc
+from werkzeug.exceptions import BadGateway, BadRequest
 
 from .. import testutil, util
 from ..util import json_dumps, json_loads
@@ -81,7 +85,7 @@ class UtilTest(testutil.TestCase):
 
     # iterator and generator
     self.assertEqual(['a', 'b'], list(util.trim_nulls(iter(['a', None, 'b']))))
-    self.assertEqual(['a', 'b'], list(util.trim_nulls(x for x in ['a', None, 'b'])))
+    self.assertEqual(['a', 'b'], list(util.trim_nulls(iter(['a', None, 'b']))))
 
     # ignore
     self.assertEqual({'a': None}, util.trim_nulls({'a': None}, ignore=('a',)))
@@ -201,6 +205,9 @@ class UtilTest(testutil.TestCase):
       self.assertEqual('asdf.com', actual, '%s returned %s' % (good_link, actual))
 
     self.assertEqual('asdf.com.', util.domain_from_link('http://asdf.com./x'))
+    self.assertEqual('⊙.de', util.domain_from_link('http://⊙.de/x'))
+    self.assertEqual('abc⊙.de', util.domain_from_link('http://abc⊙.de/x'))
+    self.assertEqual('abc.⊙.de', util.domain_from_link('http://abc.⊙.de/x'))
 
     for bad_link in '', '  ', 'a&b.com', 'http://', 'file:///':
       self.assertEqual(None, util.domain_from_link(bad_link))
@@ -222,24 +229,14 @@ class UtilTest(testutil.TestCase):
                           repr((input, domains, expected)))
 
   def test_update_scheme(self):
-    self.request = webapp2.Request.blank('/')
-    handler = webapp2.RequestHandler(self.request, None)
+    request = webapp2.Request.blank('/')
 
     for orig in 'http', 'https':
       for new in 'http', 'https':
-        handler.request.scheme = new
-        updated = util.update_scheme(orig + '://foo', handler)
+        request.scheme = new
+        updated = util.update_scheme(orig + '://foo', request)
         self.assertEqual(new + '://foo', updated)
 
-    handler.request.scheme = 'https'
-    self.assertEqual(
-      'https://distillery.s3.amazonaws.com/profiles/xyz.jpg',
-      util.update_scheme('http://images.ak.instagram.com/profiles/xyz.jpg', handler))
-    self.assertEqual(
-      'https://igcdn-photos-e-a.akamaihd.net/hphotos-ak-xpf1/123_a.jpg',
-      util.update_scheme('http://photos-e.ak.instagram.com/hphotos-ak-xpf1/123_a.jpg',
-                         handler))
-
   def test_schemeless(self):
     for expected, url in (
         ('', ''),
@@ -280,6 +277,17 @@ class UtilTest(testutil.TestCase):
     self.assertEqual('http://foo?source=not-rss',
                      util.clean_url('http://foo?&source=not-rss'))
 
+  def test_quote_path(self):
+    for unchanged in '', 'foo', 'http://foo', 'http://foo#bar', 'http://foo?x=y&z=w':
+      self.assertEqual(unchanged, util.quote_path(unchanged))
+
+    for input, expected in (
+        ('http://x/☕', 'http://x/%E2%98%95'),
+        ('http://x/foo ☕ bar', 'http://x/foo%20%E2%98%95%20bar'),
+        ('http://☕:☕@☕.com/☕?☕=☕&☕=☕#☕', 'http://☕:☕@☕.com/%E2%98%95?☕=☕&☕=☕#☕'),
+    ):
+      self.assertEqual(expected, util.quote_path(input))
+
   def test_dedupe_urls(self):
     self.assertEqual([], util.dedupe_urls([]))
     self.assertEqual([], util.dedupe_urls(['', None, '']))
@@ -359,6 +367,11 @@ class UtilTest(testutil.TestCase):
     self.assertEqual(['http://foo.com/z-'],
                       util.extract_links('x http://foo.com/z-'))
 
+    # omit trailing close parentheses (eg in markdown links)
+    self.assertEqual(['http://xy/z'], util.extract_links('[abc](http://xy/z)'))
+    self.assertEqual(['http://xy/z'], util.extract_links('[abc](http://xy/z).'))
+    self.assertEqual(['http://xy/z'], util.extract_links('([abc](http://xy/z))'))
+
     # query
     self.assertEqual(['http://foo.com/bar?baz=baj'],
                       util.extract_links('http://foo.com/bar?baz=baj y'))
@@ -373,6 +386,10 @@ class UtilTest(testutil.TestCase):
     self.assertEqual(['http://%s' % c for c in ('a', 'b', 'c', 'd')],
                       util.extract_links('http://a http://b http://c http://d'))
 
+    # emoji in domain
+    self.assertEqual(['http://☕⊙.ws'],
+                     util.extract_links('emoji http://☕⊙.ws domain'))
+
   def test_linkify(self):
     for unchanged in (
         '',
@@ -382,8 +399,6 @@ class UtilTest(testutil.TestCase):
         '<a href="http://foo.com"  class="x">xyz</a> Y',
         "X <a href='http://foo.com' />",
         'asdf <a href="http://foo.com">foo</a> qwert ',
-        # TODO: support unicode chars
-        'http://aÇb.com',
         'http://a<b>.com'):
       self.assertEqual(unchanged, util.linkify(unchanged))
 
@@ -413,9 +428,22 @@ class UtilTest(testutil.TestCase):
          'interesting how twitter auto-links it http://example.com/a_link_(with_parens) vs. (http://example.com/a_link_without)'),
         ('links separated by punctuation <a href="http://foo.com">http://foo.com</a>, <a href="http://bar.com/">http://bar.com/</a>; <a href="http://baz.com/?s=query">http://baz.com/?s=query</a>; did it work?',
          'links separated by punctuation http://foo.com, http://bar.com/; http://baz.com/?s=query; did it work?'),
+        ('"<a href="http://foo.com">http://foo.com</a>",', '"http://foo.com",'),
+        ('\'<a href="http://foo.com">http://foo.com</a>\',', "'http://foo.com',"),
+        ('<a href="http://aÇb.com">http://aÇb.com</a>', 'http://aÇb.com'),
+        ('<a href="http://a☕⊙b.com">http://a☕⊙b.com</a>', 'http://a☕⊙b.com'),
+        ('<a href="http://a☕⊙b.com">a☕⊙b.com</a>', 'a☕⊙b.com'),
+        ('<a href="http://☕⊙.ws">http://☕⊙.ws</a>', 'http://☕⊙.ws'),
+        # TODO: implement
+        # see comments in regexps at top of util.py for details
+        # ('<a href="http://☕⊙.ws">☕⊙.ws</a>', '☕⊙.ws'),
     ):
         self.assertEqual(expected, util.linkify(input))
 
+    # test skip_bare_cc_tlds
+    for unchanged in ('x ab.ws y', 'x ☕⊙.ws y'):
+      self.assertEqual(unchanged, util.linkify(unchanged, skip_bare_cc_tlds=True))
+
   def test_pretty_link(self):
     pl = util.pretty_link
     self.assertEqual('<a href="http://foo">foo</a>', pl('http://foo'))
@@ -476,13 +504,19 @@ class UtilTest(testutil.TestCase):
     self.assertEqual('<a href="http://x/ben-werdmüller">x/ben-werdmüller</a>',
                      pl('http://x/ben-werdmüller'))
 
+    self.assertEqual('<a href="http://aÇb.com">aÇb.com</a>', pl('http://aÇb.com'))
+    self.assertEqual('<a href="http://a☕⊙b.com">a☕⊙b.com</a>',
+                     pl('http://a☕⊙b.com'))
+
   # TODO: make this work
   # def test_linkify_broken(self):
   #   self.assertEqual('', util.linkify(
   #       '<a href="http://www.example.com/?feature_embedded">'))
 
   def test_linkify_pretty(self):
-    lp = lambda url: util.linkify(url, pretty=True, max_length=6)
+    def lp(url):
+      return util.linkify(url, pretty=True, max_length=6)
+
     self.assertEqual('', lp(''))
     self.assertEqual('asdf qwert', lp('asdf qwert'))
     self.assertEqual('x <a href="http://foo.co">foo.co</a> y', lp('x http://foo.co y'))
@@ -562,6 +596,18 @@ class UtilTest(testutil.TestCase):
     ):
       self.assertEqual(expected, util.maybe_timestamp_to_rfc3339(input))
 
+  def test_maybe_timestamp_to_iso8601(self):
+    for input, expected in (
+      (None, None),
+      ('', ''),
+      ('not a timestamp!', 'not a timestamp!'),
+      (1349588757, '2012-10-07T05:45:57Z'),
+      ('1349588757', '2012-10-07T05:45:57Z'),
+      (1349588757.123, '2012-10-07T05:45:57.123Z'),
+      (1349588757.123456, '2012-10-07T05:45:57.123Z'),
+    ):
+      self.assertEqual(expected, util.maybe_timestamp_to_iso8601(input))
+
   def test_to_utc_timestamp(self):
     self.assertIsNone(util.to_utc_timestamp(None))
     self.assertEqual(0, util.to_utc_timestamp(datetime.datetime(1970, 1, 1)))
@@ -602,7 +648,7 @@ class UtilTest(testutil.TestCase):
       # note encoding declaration at top of file
       ('http://a.com?x=R+%C3%87', 'http://a.com', {'x': 'R Ç'}),
       ('http://a.com?x=R+%C3%87&x=R+%C3%87', 'http://a.com?x=R+%C3%87', {'x': 'R Ç'}),
-      ):
+    ):
       self.assertEqual(expected, util.add_query_params(url, params))
 
     for expected, req, params in (
@@ -613,7 +659,7 @@ class UtilTest(testutil.TestCase):
       (urllib.request.Request('http://a.com?x=y', data='my data', headers={'X': 'Y'}),
        urllib.request.Request('http://a.com', data='my data', headers={'X': 'Y'}),
        [('x', 'y')]),
-      ):
+    ):
       actual = util.add_query_params(req, params)
       self.assertIsInstance(actual, urllib.request.Request)
       self.assertEqual(expected.get_full_url(), actual.get_full_url())
@@ -621,7 +667,7 @@ class UtilTest(testutil.TestCase):
       self.assertEqual(expected.headers, actual.headers)
 
     query_string = ''
-    for i in range(2):
+    for _ in range(2):
       query_string = util.add_query_params(query_string, {'x': 'Ryan Çelik'})
       for key, val in urllib.parse.parse_qsl(query_string[1:]):
         self.assertEqual('x', key)
@@ -690,7 +736,7 @@ class UtilTest(testutil.TestCase):
     self.assert_equals(data, cd.get_multi({1, 3}))
 
     # get_multi should handle a generator args ok
-    self.assert_equals(data, cd.get_multi(k for k in [1, 3]))
+    self.assert_equals(data, cd.get_multi(iter([1, 3])))
     self.assert_equals(data, cd.get_multi(list(range(4))))
 
   def test_is_int(self):
@@ -717,6 +763,8 @@ class UtilTest(testutil.TestCase):
     self.assertEqual(('402', '402 Payment Required\n\nmy body'), ihc(
         exc.HTTPPaymentRequired(body_template='my body')))
 
+    self.assertEqual(('502', '<p>my body</p>'), ihc(BadGateway('my body')))
+
     # rate limiting
     ex = HTTPError('url', 429, 'msg', {}, io.StringIO('my body'))
     self.assertFalse(util.is_connection_failure(ex))
@@ -730,7 +778,9 @@ class UtilTest(testutil.TestCase):
         requests.HTTPError(response=util.Struct(status_code='429', text='my body'))))
 
     # facebook page rate limiting
-    json_str = lambda obj: io.StringIO(str(json_dumps(obj)))
+    def json_str(obj):
+      return io.StringIO(str(json_dumps(obj)))
+
     body = {'error': {
       'type': 'OAuthException',
       'code': 32,
@@ -742,9 +792,22 @@ class UtilTest(testutil.TestCase):
     # fake gdata.client.RequestError since gdata isn't a dependency
     class RequestError(util.Struct):
       pass
+
     ex = RequestError(status=429, body='my body')
     self.assertEqual(('429', 'my body'), ihc(ex))
 
+    ex = RequestError(status=429, body=b'my body')
+    self.assertEqual(('429', 'my body'), ihc(ex))
+
+    ex = tumblpy.TumblpyError('my body', error_code=429)
+    self.assertEqual(('429', 'my body'), ihc(ex))
+
+    ex = tweepy.TweepError('my body')
+    self.assertEqual(('400', 'my body'), ihc(ex))
+
+    ex = tweepy.RateLimitError('my body')
+    self.assertEqual(('429', 'my body'), ihc(ex))
+
     # Flickr
     #
     # generated by oauth_dropins.flickr_auth.raise_for_failure()
@@ -764,10 +827,10 @@ class UtilTest(testutil.TestCase):
         'error_message': 'The access_token provided is invalid.'
       }},
       # facebook, https://github.com/snarfed/bridgy/issues/59#issuecomment-34549314
-      {'error' : {
-        'type' : 'OAuthException',
-        'code' : 100,
-        'message' : 'This authorization code has expired.'
+      {'error': {
+        'type': 'OAuthException',
+        'code': 100,
+        'message': 'This authorization code has expired.'
       }},
       # facebook, https://github.com/snarfed/bridgy/issues/436
       {'error': {
@@ -848,15 +911,15 @@ class UtilTest(testutil.TestCase):
       }},
       # facebook deprecated API, https://github.com/snarfed/bridgy/issues/480
       {'error': {
-        'message' : '(#12) notes API is deprecated for versions v2.0 and higher',
-        'code' : 12,
-        'type' : 'OAuthException',
+        'message': '(#12) notes API is deprecated for versions v2.0 and higher',
+        'code': 12,
+        'type': 'OAuthException',
       }},
       # facebook too many IDs
       {'error': {
-        'message' : '(#100) Too many IDs. Maximum: 50. Provided: 54.',
-        'code' : 100,
-        'type' : 'OAuthException',
+        'message': '(#100) Too many IDs. Maximum: 50. Provided: 54.',
+        'code': 100,
+        'type': 'OAuthException',
       }},
     ):
       for code, expected in (400, 400), (500, 502):
@@ -868,11 +931,11 @@ class UtilTest(testutil.TestCase):
     # facebook temporarily unavailable with is_transient
     # https://github.com/snarfed/bridgy/issues/450
     fb_transient = {'error': {
-      'message' : '(#2) Service temporarily unavailable',
+      'message': '(#2) Service temporarily unavailable',
       # also seen:
       # 'message': 'An unexpected error has occurred. Please retry your request later.',
-      'code' : 2,
-      'type' : 'OAuthException',
+      'code': 2,
+      'type': 'OAuthException',
       'is_transient': True,
     }}
     for code in (400, 500):
@@ -881,6 +944,12 @@ class UtilTest(testutil.TestCase):
         ('503', body.getvalue()),
         ihc(HTTPError('url', code, 'BAD REQUEST', {}, body)))
 
+    # pleroma JSON string error message
+    # https://console.cloud.google.com/errors/CP3p1f2g_MeF5AE
+    code, body = ihc(HTTPError('url', 500, '', {}, json_str('Something went wrong')))
+    self.assertEqual('502', code)
+    self.assertEqual('"Something went wrong"', body)
+
     # make sure we handle non-facebook JSON bodies ok
     wordpress_rest_error = json_str({
       'error': 'unauthorized',
@@ -899,6 +968,10 @@ class UtilTest(testutil.TestCase):
     # upstream connection failures are converted to 504
     self.assertEqual(('504', 'foo bar'), ihc(socket.timeout('foo bar')))
 
+    self.assertEqual(('504', 'foo bar'),
+                     ihc(prawcore.exceptions.ResponseException(
+                       testutil.requests_response('foo bar', status=504))))
+
   def test_ignore_http_4xx_error(self):
     x = 0
     with util.ignore_http_4xx_error():
@@ -923,6 +996,7 @@ class UtilTest(testutil.TestCase):
         urllib3.exceptions.TimeoutError(),
         Exception('Connection closed unexpectedly by server at URL: ...'),
         ssl.SSLError(),
+        prawcore.exceptions.RequestException(None, None, None),
     ):
       self.assertTrue(util.is_connection_failure(e), repr(e))
 
@@ -969,7 +1043,7 @@ class UtilTest(testutil.TestCase):
       self.assertEqual(expected, util.base_url(url))
 
   def test_follow_redirects(self):
-    for i in range(2):
+    for _ in range(2):
       self.expect_requests_head('http://will/redirect',
                                 redirected_url='http://final/url')
     self.mox.ReplayAll()
@@ -1155,7 +1229,7 @@ class UtilTest(testutil.TestCase):
     self.mox.ReplayAll()
 
     with self.assertRaises(requests.HTTPError) as e:
-      resp = util.requests_post_with_redirects('http://first')
+      util.requests_post_with_redirects('http://first')
 
     self.assert_equals('abc', e.exception.response.text)
     self.assert_equals(400, e.exception.response.status_code)
@@ -1168,20 +1242,20 @@ class UtilTest(testutil.TestCase):
     self.mox.ReplayAll()
 
     with self.assertRaises(requests.HTTPError) as e:
-      resp = util.requests_post_with_redirects('http://ok')
+      util.requests_post_with_redirects('http://ok')
 
     self.assert_equals('abc', e.exception.response.text)
     self.assert_equals(400, e.exception.response.status_code)
 
   def test_requests_post_with_redirects_too_many_redirects(self):
-    for i in range(requests.models.DEFAULT_REDIRECT_LIMIT):
+    for _ in range(requests.models.DEFAULT_REDIRECT_LIMIT):
       self.expect_requests_post(
         'http://xyz', 'abc', allow_redirects=False, status_code=302,
         response_headers={'Location': 'http://xyz'})
     self.mox.ReplayAll()
 
     with self.assertRaises(requests.TooManyRedirects) as e:
-      resp = util.requests_post_with_redirects('http://xyz')
+      util.requests_post_with_redirects('http://xyz')
 
     self.assert_equals('abc', e.exception.response.text)
     self.assert_equals(302, e.exception.response.status_code)
@@ -1198,7 +1272,7 @@ class UtilTest(testutil.TestCase):
     for i in range(2):
       self.assertEqual(422, util.requests_get('http://xyz').status_code, i)
 
-    for i in range(2):
+    for _ in range(2):
       self.assertEqual(200, util.requests_get('http://xyz').status_code)
 
   def test_requests_get_unicode_url_ValueError(self):
@@ -1206,11 +1280,21 @@ class UtilTest(testutil.TestCase):
     url = 'http://acct:abc⊙de/'
     self.expect_requests_get(url).AndRaise(ValueError())
     self.mox.ReplayAll()
-    self.assertRaises(exc.HTTPBadRequest, util.requests_get, url, gateway=True)
+    self.assertRaises(BadRequest, util.requests_get, url, gateway=True)
 
   def test_requests_get_unicode_url_ConnectionError(self):
     """https://console.cloud.google.com/errors/CPzNwYaL3tjb9gE"""
     url = 'http://acct:abc⊙de/'
     self.expect_requests_get(url).AndRaise(requests.ConnectionError())
     self.mox.ReplayAll()
-    self.assertRaises(exc.HTTPBadGateway, util.requests_get, url, gateway=True)
+    self.assertRaises(BadGateway, util.requests_get, url, gateway=True)
+
+  def test_requests_get_invalid_emoji_domain_fallback_to_domain2idnaError(self):
+    url = 'http://abc⊙.de/'
+    self.expect_requests_get(url).AndRaise(requests.exceptions.InvalidURL())
+    self.expect_requests_get('http://xn--abc-yr2a.de/', 'ok')
+    self.mox.ReplayAll()
+
+    resp = util.requests_get(url)
+    self.assertEqual(200, resp.status_code)
+    self.assertEqual(url, resp.url)
