@@ -1,11 +1,11 @@
 """Misc utilities.
 
-Supports Python 3. Should not depend on App Engine API or SDK packages.
+Should not depend on App Engine API or SDK packages.
 """
 import calendar
 import collections
+from collections.abc import Iterator
 import contextlib
-import copy
 import base64
 import datetime
 import http.client
@@ -17,11 +17,17 @@ import os
 import re
 import socket
 import ssl
+import string
+import sys
 import threading
+import traceback
 import urllib.error, urllib.parse, urllib.request
+from urllib.parse import urlparse
 from xml.sax import saxutils
 
 from cachetools import cached, TTLCache
+from domain2idna import domain2idna
+from flask import abort
 
 try:
   import ujson
@@ -69,6 +75,12 @@ try:
 except ImportError:
   exc = None
 
+try:
+  import werkzeug
+  import werkzeug.exceptions
+except ImportError:
+  werkzeug = None
+
 # Used in parse_html() and friends.
 try:
   import bs4
@@ -80,6 +92,21 @@ try:
 except ImportError:
   mf2py = None
 
+try:
+  import prawcore
+except ImportError:
+  prawcore = None
+
+try:
+  import tumblpy
+except ImportError:
+  tumblpy = None
+
+try:
+  import tweepy
+except ImportError:
+  tweepy = None
+
 EPOCH = datetime.datetime.utcfromtimestamp(0)
 EPOCH_ISO = EPOCH.isoformat()
 # from https://stackoverflow.com/a/53140944/186123
@@ -88,7 +115,6 @@ ISO8601_DURATION_RE = re.compile(
 
 # default HTTP request timeout
 HTTP_TIMEOUT = 15
-import socket
 socket.setdefaulttimeout(HTTP_TIMEOUT)
 # monkey-patch socket.getdefaulttimeout() because it often gets reset, e.g. by
 # socket.setblocking() and maybe other operations.
@@ -106,12 +132,64 @@ FOLLOW_REDIRECTS_CACHE_TIME = 60 * 60 * 24  # 1d expiration
 follow_redirects_cache = TTLCache(1000, FOLLOW_REDIRECTS_CACHE_TIME)
 follow_redirects_cache_lock = threading.RLock()
 
+# https://en.wikipedia.org/wiki/Top-level_domain#Reserved_domains
+# Currently used in granary.source.Source.original_post_discovery, not here.
+RESERVED_TLDS = {
+  'corp',
+  'example',
+  'internal',
+  'invalid',
+  'onion',
+  'test',
+}
+LOCAL_TLDS = {
+  'local',
+  'localhost',
+}
+
 """Global config, string parser nae for BeautifulSoup to use, e.g. 'lxml'.
 May be set at runtime.
 https://www.crummy.com/software/BeautifulSoup/bs4/doc/#installing-a-parser
 """
 beautifulsoup_parser = None
 
+# Regexps for domains, hostnames, and URLs.
+#
+# Based on kylewm's from redwind:
+# https://github.com/snarfed/bridgy/issues/209#issuecomment-47583528
+# https://github.com/kylewm/redwind/blob/863989d48b97a85a1c1a92c6d79753d2fbb70775/redwind/util.py#L39
+#
+# I used to use a more complicated regexp based on
+# https://github.com/silas/huck/blob/master/huck/utils.py#L59 , but i kept
+# finding new input strings that would make it hang the regexp engine.
+#
+# more complicated alternatives:
+# http://stackoverflow.com/questions/720113#comment23297770_2102648
+# https://daringfireball.net/2010/07/improved_regex_for_matching_urls
+#
+# list of TLDs:
+# https://en.wikipedia.org/wiki/List_of_Internet_top-level_domains#ICANN-era_generic_top-level_domains
+#
+# Allows emoji and other unicode chars in all domain labels *except* TLDs.
+# TODO: support IDN TLDs:
+# https://en.wikipedia.org/wiki/Top-level_domain#Internationalized_country_code_TLDs
+# https://www.iana.org/domains/root/db
+#
+# TODO: fix bug in LINK_RE that makes it miss emoji domain links without scheme,
+# eg '☕⊙.ws'. bug is that the \b at the beginning of SCHEME_RE doesn't apply to
+# emoji, since they're not word-constituent characters, and that the '?' added
+# in LINK_RE only applies to the parenthesized group in SCHEME_RE, not the \b.
+# I tried changing \b to '(?:^|[\s%s])' % PUNCT, but that broke other things.
+PUNCT = string.punctuation.replace('-', '').replace('.', '')
+SCHEME_RE = r'\b(?:[a-z]{3,9}:/{1,3})'
+HOST_RE = r'(?:[^\s%s])+(?::\d{2,6})?' % PUNCT
+DOMAIN_RE = r'(?:[^\s.%s]+\.)+[a-z]{2,}(?::\d{2,6})?' % PUNCT
+PATH_QUERY_RE = r'(?:(?:/[\w/.\-_~.;:%?@$#&()=+]*)|\b)'
+URL_RE = re.compile(SCHEME_RE + HOST_RE + PATH_QUERY_RE,  # scheme required
+                    re.UNICODE | re.IGNORECASE)
+LINK_RE = re.compile(SCHEME_RE + '?' + DOMAIN_RE + PATH_QUERY_RE,  # scheme optional
+                     re.UNICODE | re.IGNORECASE)
+
 
 class Struct(object):
   """A generic class that initializes its attributes from constructor kwargs."""
@@ -165,11 +243,11 @@ def trim_nulls(value, ignore=()):
   if isinstance(value, dict):
     trimmed = {k: trim_nulls(v, ignore=ignore) for k, v in value.items()}
     return {k: v for k, v in trimmed.items() if k in ignore or v not in NULLS}
-  elif (isinstance(value, (tuple, list, set, frozenset, collections.Iterator)) or
+  elif (isinstance(value, (tuple, list, set, frozenset, Iterator)) or
         inspect.isgenerator(value)):
     trimmed = [trim_nulls(v, ignore=ignore) for v in value]
     ret = (v for v in trimmed if v if v not in NULLS)
-    if isinstance(value, collections.Iterator) or inspect.isgenerator(value):
+    if isinstance(value, Iterator) or inspect.isgenerator(value):
       return ret
     else:
       return type(value)(list(ret))
@@ -268,14 +346,11 @@ def tag_uri(domain, name, year=None):
 
   Background on tag URIs: http://taguri.org/
   """
-  if year is not None:
-    year = ',%s' % year
-  else:
-    year = ''
-  return 'tag:%s%s:%s' % (domain, year, name)
+  year = ',%s' % year if year else ''
+  return f'tag:{domain}{year}:{name}'
 
 
-_TAG_URI_RE = re.compile('tag:([^,]+)(?:,\d+)?:(.+)$')
+_TAG_URI_RE = re.compile(r'tag:([^,]+)(?:,\d+)?:(.+)$')
 
 def parse_tag_uri(uri):
   """Returns the domain and name in a tag URI string.
@@ -304,7 +379,7 @@ def parse_acct_uri(uri, hosts=None):
 
   Raises: ValueError if the uri is invalid or the host isn't allowed.
   """
-  parsed = urllib.parse.urlparse(uri)
+  parsed = urlparse(uri)
   if parsed.scheme and parsed.scheme != 'acct':
     raise ValueError('Acct URI %s has unsupported scheme: %s' %
                      (uri, parsed.scheme))
@@ -323,12 +398,10 @@ def parse_acct_uri(uri, hosts=None):
 
 
 def favicon_for_url(url):
-  return 'http://%s/favicon.ico' % urllib.parse.urlparse(url).netloc
+  return 'http://%s/favicon.ico' % urlparse(url).netloc
 
 
-# http://stackoverflow.com/questions/2532053/validate-hostname-string-in-python
-HOSTNAME_RE_STR = r'%s(\.%s)*\.?' % ((r'(?!-)[A-Za-z\d-]{1,63}(?<!-)',) * 2)
-HOSTNAME_RE = re.compile(HOSTNAME_RE_STR + '$')
+FULL_HOST_RE = re.compile(HOST_RE + '$')
 
 def domain_from_link(url):
   """Extracts and returns the meaningful domain from a URL.
@@ -341,17 +414,17 @@ def domain_from_link(url):
   Returns:
     string
   """
-  parsed = urllib.parse.urlparse(url)
+  parsed = urlparse(url)
   if not parsed.hostname and '//' not in url:
-    parsed = urllib.parse.urlparse('http://' + url)
+    parsed = urlparse('http://' + url)
 
   domain = parsed.hostname
   if domain:
     for subdomain in ('www.', 'mobile.', 'm.'):
       if domain.startswith(subdomain):
         domain = domain[len(subdomain):]
-    if domain and HOSTNAME_RE.match(domain):
-      return domain
+  if domain and FULL_HOST_RE.match(domain):
+    return domain
 
   return None
 
@@ -364,6 +437,7 @@ def domain_or_parent_in(input, domains):
   * foo, [] => False
   * foo, [foo] => True
   * foo.bar.com, [bar.com] => True
+  * foobar.com, [bar.com] => False
   * foo.bar.com, [.bar.com] => True
   * foo.bar.com, [fux.bar.com] => False
   * bar.com, [fux.bar.com] => False
@@ -388,7 +462,8 @@ def domain_or_parent_in(input, domains):
 
   return False
 
-def update_scheme(url, handler):
+
+def update_scheme(url, request):
   """Returns a modified URL with the current request's scheme.
 
   Useful for converting URLs to https if and only if the current request itself
@@ -396,20 +471,11 @@ def update_scheme(url, handler):
 
   Args:
     url: string
-    handler: :class:`webapp2.RequestHandler`
+    request: :class:`flask.Request` or :class:`webob.Request`
 
   Returns: string, url
   """
-  # Instagram can't serve images over SSL, so switch to their S3 or Akamai URLs,
-  # which can.
-  # https://groups.google.com/d/msg/instagram-api-developers/fB4mwYXZF1c/q9n9gPO11JQJ
-  # http://stackoverflow.com/questions/23755897#comment36547383_23755897
-  url = re.sub(r'^http://images\.(ak\.)instagram\.com',
-               'http://distillery.s3.amazonaws.com', url)
-  url = re.sub(r'^http://photos-\w\.(ak\.)instagram\.com',
-               'http://igcdn-photos-e-a.akamaihd.net', url)
-  return urllib.parse.urlunparse((handler.request.scheme,) +
-                                 urllib.parse.urlparse(url)[1:])
+  return urllib.parse.urlunparse((request.scheme,) + urlparse(url)[1:])
 
 
 def schemeless(url, slashes=True):
@@ -423,7 +489,7 @@ def schemeless(url, slashes=True):
   Returns:
     string URL
   """
-  url = urllib.parse.urlunparse(('',) + urllib.parse.urlparse(url)[1:])
+  url = urllib.parse.urlunparse(('',) + urlparse(url)[1:])
   if not slashes:
     url = url.strip('/')
   return url
@@ -438,7 +504,7 @@ def fragmentless(url):
   Returns:
     string URL
   """
-  return urllib.parse.urlunparse(urllib.parse.urlparse(url)[:5] + ('',))
+  return urllib.parse.urlunparse(urlparse(url)[:5] + ('',))
 
 
 def clean_url(url):
@@ -461,7 +527,7 @@ def clean_url(url):
   utm_params = set(('utm_campaign', 'utm_content', 'utm_medium', 'utm_source',
                     'utm_term'))
   try:
-    parts = list(urllib.parse.urlparse(url))
+    parts = list(urlparse(url))
   except (AttributeError, TypeError, ValueError):
     return None
 
@@ -473,6 +539,24 @@ def clean_url(url):
   return urllib.parse.urlunparse(parts)
 
 
+def quote_path(url):
+  """Quotes (URL-encodes) just the path part of a URL.
+
+  Args:
+    url: string
+
+  Returns:
+    string, the quoted url, or None if it can't be parsed
+  """
+  try:
+    parts = list(urlparse(url))
+  except (AttributeError, TypeError, ValueError):
+    return None
+
+  parts[2] = urllib.parse.quote(parts[2])
+  return urllib.parse.urlunparse(parts)
+
+
 def base_url(url):
   """Returns the base of a given URL.
 
@@ -484,33 +568,6 @@ def base_url(url):
   return urllib.parse.urljoin(url, ' ')[:-1] if url else None
 
 
-# Based on kylewm's from redwind:
-# https://github.com/snarfed/bridgy/issues/209#issuecomment-47583528
-# https://github.com/kylewm/redwind/blob/863989d48b97a85a1c1a92c6d79753d2fbb70775/redwind/util.py#L39
-#
-# I used to use a more complicated regexp based on
-# https://github.com/silas/huck/blob/master/huck/utils.py#L59 , but i kept
-# finding new input strings that would make it hang the regexp engine.
-#
-# more complicated alternative:
-# http://stackoverflow.com/questions/720113#comment23297770_2102648
-#
-# list of TLDs:
-# https://en.wikipedia.org/wiki/List_of_Internet_top-level_domains#ICANN-era_generic_top-level_domains
-#
-# TODO: support internationalized/emoji TLDs
-_SCHEME_RE = r'\b(?:[a-z]{3,9}:/{1,3})'
-_HOST_RE =   r'(?:[a-z0-9\-.])+(?::\d{2,6})?'
-_DOMAIN_RE = r'(?:[a-z0-9\-]+\.)+[a-z0-9\-]{2,}(?::\d{2,6})?'
-_PATH_QUERY_RE = r'(?:(?:/[\w/.\-_~.;:%?@$#&()=+]*)|\b)'
-_LINK_RE = re.compile(
-  _SCHEME_RE + _HOST_RE + _PATH_QUERY_RE,  # scheme required
-  re.UNICODE | re.IGNORECASE)
-_LINKIFY_RE = re.compile(
-  _SCHEME_RE + '?' + _DOMAIN_RE + _PATH_QUERY_RE,  # scheme optional
-  re.UNICODE | re.IGNORECASE)
-
-
 def extract_links(text):
   """Returns a list of unique string URLs in the given text.
 
@@ -519,28 +576,29 @@ def extract_links(text):
   if not text:
     return []
 
-  links = uniquify(match.group() for match in _LINK_RE.finditer(text))
-  # strip "outside" parens
-  links = [l[:-1] if l[-1] == ')' and '(' not in l else l
-           for l in links]
-  return links
+  return uniquify(tokenize_links(text, skip_html_links=False, require_scheme=True)[0])
 
 
-def tokenize_links(text, skip_bare_cc_tlds=False):
+def tokenize_links(text, skip_bare_cc_tlds=False, skip_html_links=True,
+                   require_scheme=False):
   """Splits text into link and non-link text.
 
   Args:
     text: string to linkify
     skip_bare_cc_tlds: boolean, whether to skip links of the form
       [domain].[2-letter TLD] with no schema and no path
+    skip_html_links: boolean, whether to skip links in HTML <a> tags (
+      both href and text)
+    require_scheme: boolean, whether to require scheme (eg http:// )
 
   Returns:
     a tuple containing two lists of strings, a list of links and list of
     non-link text. Roughly equivalent to the output of re.findall and re.split,
     with some post-processing.
   """
-  links = _LINKIFY_RE.findall(text)
-  splits = _LINKIFY_RE.split(text)
+  regexp = URL_RE if require_scheme else LINK_RE
+  links = regexp.findall(text)
+  splits = regexp.split(text)
 
   for ii in range(len(links)):
     # trim trailing punctuation from links
@@ -556,11 +614,11 @@ def tokenize_links(text, skip_bare_cc_tlds=False):
     link = links[ii]
 
     # avoid double linking by looking at preceeding 2 chars
-    if (splits[ii].strip().endswith('="')
-        or splits[ii].strip().endswith("='")
-        or splits[ii + 1].strip().startswith('</a')
+    if ((skip_html_links and (splits[ii].strip().endswith('="')
+                              or splits[ii].strip().endswith("='")
+                              or splits[ii + 1].strip().startswith('</a')))
         # skip domains with 2-letter TLDs and no schema or path
-        or (skip_bare_cc_tlds and re.match('[a-z0-9\-]+\.[a-z]{2}$', link))):
+        or (skip_bare_cc_tlds and re.match(r'[^\s%s]+\.[a-z]{2}$' % PUNCT, link))):
       # collapse link into before text
       splits[ii] = splits[ii] + links[ii]
       links[ii] = None
@@ -647,7 +705,7 @@ def pretty_link(url, text=None, keep_host=True, glyphicon=None, attrs=None,
       max_length = 30
   else:
     # use shortened version of URL as link text
-    parsed = urllib.parse.urlparse(url)
+    parsed = urlparse(url)
     text = url[len(parsed.scheme) + 3:]  # strip scheme and ://
     host_len = len(parsed.netloc)
     if (keep_host and not parsed.params and not parsed.query and not parsed.fragment):
@@ -824,6 +882,15 @@ def maybe_timestamp_to_rfc3339(input):
     return input
 
 
+def maybe_timestamp_to_iso8601(input):
+  """Tries to convert a string or int UNIX timestamp to ISO 8601.
+
+  Assumes UNIX timestamps are always UTC. (They're generally supposed to be.)
+  """
+  ret = maybe_timestamp_to_rfc3339(input)
+  return ret if ret == input else ret.replace('+00:00', 'Z')
+
+
 def to_utc_timestamp(input):
   """Converts a datetime to a float POSIX timestamp (seconds since epoch)."""
   if not input:
@@ -842,12 +909,12 @@ def as_utc(input):
 
   Doesn't support DST!
   """
-  if input.tzinfo:
-    utc = input - input.tzinfo.utcoffset(False)
-    return utc.replace(tzinfo=None)
-  else:
+  if not input.tzinfo:
     return input
 
+  utc = input - input.tzinfo.utcoffset(False)
+  return utc.replace(tzinfo=None)
+
 
 def ellipsize(str, words=14, chars=140):
   """Truncates and ellipsizes str if it's longer than words or chars.
@@ -857,7 +924,7 @@ def ellipsize(str, words=14, chars=140):
   split = str.split()
   if len(split) <= words and len(str) <= chars:
     return str
-  return ' '.join(split[:words])[:chars-3] + '...'
+  return ' '.join(split[:words])[:chars - 3] + '...'
 
 
 def add_query_params(url, params):
@@ -880,7 +947,7 @@ def add_query_params(url, params):
     params = list(params.items())
 
   # convert to list so we can modify later
-  parsed = list(urllib.parse.urlparse(url))
+  parsed = list(urlparse(url))
   # query params are in index 4
   params = [(k, str(v).encode('utf-8')) for k, v in params]
   parsed[4] += ('&' if parsed[4] else '') + urllib.parse.urlencode(params)
@@ -909,7 +976,7 @@ def remove_query_param(url, param):
     (string URL without the given param, string param value)
   """
   # convert to list so we can modify later
-  parsed = list(urllib.parse.urlparse(url))
+  parsed = list(urlparse(url))
 
   # query params are in index 4
   removed = None
@@ -924,6 +991,7 @@ def remove_query_param(url, param):
   url = urllib.parse.urlunparse(parsed)
   return url, removed
 
+
 def get_required_param(handler, name):
   try:
     val = handler.request.get(name)
@@ -1030,8 +1098,8 @@ def decode_oauth_state(state):
   try:
     obj = json_loads(urllib.parse.unquote_plus(state)) if state else {}
   except ValueError:
-    logging.error('Invalid value for state parameter: %s' % state, stack_info=True)
-    raise exc.HTTPBadRequest('Invalid value for state parameter: %s' % state)
+    logging.error(f'Invalid value for state parameter: {state}', stack_info=True)
+    abort(400, f'Invalid value for state parameter: {state}')
 
   if not isinstance(obj, dict):
     logging.error('got a non-dict state parameter %s', state)
@@ -1146,6 +1214,7 @@ def interpret_http_exception(exception):
       * :class:`requests.HTTPError`
       * :class:`urllib.error.HTTPError`
       * :class:`urllib.error.URLError`
+      * :class:`werkzeug.exceptions.HTTPException`
 
   Returns:
     (string status code or None, string response body or None)
@@ -1157,6 +1226,10 @@ def interpret_http_exception(exception):
     code = e.code
     body = e.plain_body({})
 
+  elif werkzeug and isinstance(e, werkzeug.exceptions.HTTPException):
+    code = e.code
+    body = e.get_description()
+
   elif isinstance(e, urllib.error.HTTPError):
     code = e.code
     try:
@@ -1179,10 +1252,19 @@ def interpret_http_exception(exception):
   elif isinstance(e, urllib.error.URLError):
     body = str(e.reason)
 
-  elif requests and isinstance(e, requests.HTTPError):
+  elif ((requests and isinstance(e, requests.HTTPError)) or
+        (prawcore and isinstance(e, prawcore.exceptions.ResponseException))):
     code = e.response.status_code
     body = e.response.text
 
+  elif tumblpy and isinstance(e, tumblpy.TumblpyError):
+    code = e.error_code
+    body = e.msg
+
+  elif tweepy and isinstance(e, tweepy.TweepError):
+    code = '429' if isinstance(e, tweepy.RateLimitError) else '400'
+    body = e.reason
+
   elif apiclient and isinstance(e, apiclient.errors.HttpError):
     code = e.resp.status
     body = e.content.decode('utf-8')
@@ -1208,6 +1290,13 @@ def interpret_http_exception(exception):
   if code or body:
     logging.warning('Error %s, response body: %s', code, repr(body))
 
+  if isinstance(body, bytes):
+    # good faith effort to decode as UTF-8 or ASCII
+    try:
+      body = body.decode()
+    except:
+      pass
+
   # silo-specific error_types that should disable the source.
   #
   # instagram
@@ -1222,7 +1311,7 @@ def interpret_http_exception(exception):
   if body:
     try:
       body_json = json_loads(body)
-      error = body_json.get('error', {})
+      error = body_json if isinstance(body_json, str) else body_json.get('error', {})
       if not isinstance(error, dict):
         error = {'message': repr(error)}
     except BaseException:
@@ -1254,7 +1343,7 @@ def interpret_http_exception(exception):
         # this one below comes with HTTP 400, but actually seems to be transient.
         # 'Cannot call API on behalf of this user' in message or
         'Permissions error' == message
-       )) or
+      )) or
       (type == 'FacebookApiException' and 'Permissions error' in message) or
       # https://developers.facebook.com/docs/graph-api/using-graph-api#errorcodes
       # https://developers.facebook.com/docs/graph-api/using-graph-api#errorsubcodes
@@ -1308,6 +1397,7 @@ def is_connection_failure(exception):
       http.client.ImproperConnectionState,
       http.client.IncompleteRead,
       http.client.NotConnected,
+      prawcore.exceptions.RequestException,
       socket.timeout,
       ssl.SSLError,
   ]
@@ -1329,7 +1419,9 @@ def is_connection_failure(exception):
        isinstance(exception.reason, socket.error)) or
       (isinstance(exception, http.client.HTTPException) and
        'Deadline exceeded' in msg) or
-      'Connection closed unexpectedly' in msg  # tweepy.TweepError
+      # these are tweepy.TweepError wrapping NewConnectionError
+      'Connection closed unexpectedly' in msg or
+      'Max retries exceeded' in msg
      ):
     # TODO: exc_info might not be for exception, e.g. if the json_loads() in
     # interpret_http_exception() fails. need to pass through the whole
@@ -1437,10 +1529,10 @@ def requests_fn(fn):
   Args:
     fn: 'get', 'head', or 'post'
     gateway: boolean, whether this is in a HTTP gateway request handler context.
-      If True, errors will be raised as appropriate webob HTTP exceptions.
-      Specifically, malformed URLs result in :class:`exc.HTTPBadRequest`
+      If True, errors will be raised as appropriate Flask HTTP exceptions.
+      Malformed URLs result in :class:`werkzeug.exceptions.BadRequest`
       (HTTP 400), connection failures and HTTP 4xx and 5xx result in
-      :class:`exc.HTTPBadGateway` (HTTP 502).
+      :class:`werkzeug.exceptions.BadGateway` (HTTP 502).
   """
   def call(url, *args, **kwargs):
     logging.info('requests.%s %s %s', fn, url, _prune(kwargs))
@@ -1454,17 +1546,40 @@ def requests_fn(fn):
       # use getattr so that stubbing out with mox still works
       resp = getattr(requests, fn)(url, *args, **kwargs)
       if gateway:
+        logging.info(f'Received {resp.status_code}: {"" if resp.ok else resp.text[:500]}')
         resp.raise_for_status()
     except (ValueError, requests.URLRequired) as e:
+      if isinstance(e, requests.exceptions.InvalidURL):
+        punycode = domain2idna(url)  # surprisingly, this handles full URLs fine
+        if punycode != url:
+          # the domain is valid idn2003 but not idn2008. encode and try again.
+          # https://unicode.org/faq/idn.html#6
+          # https://github.com/psf/requests/issues/3687
+          # https://github.com/kjd/idna/issues/18
+          # https://github.com/kjd/idna/issues/40
+          resp = call(punycode, *args, **kwargs)
+          resp.url = resp.url.replace(urlparse(punycode).netloc,
+                                      urlparse(url).netloc)
+          return resp
+
       if gateway:
-        msg = 'Bad URL %s' % url
-        logging.warning(msg, stack_info=True)
-        raise exc.HTTPBadRequest(msg)
+        msg = f'Bad URL {url} : {e}'
+        logging.warning(msg)
+        # this format_exc with tb None below, instead of passing exc_info=True
+        # above, prevents the 'Traceback (most recent call last):' prefix that
+        # triggers Stackdriver Error Reporting
+        logging.warning('\n'.join(traceback.format_tb(sys.exc_info()[2])))
+        abort(400, msg)
       raise
+
     except requests.RequestException as e:
       if gateway:
-        logging.warning(url, stack_info=True)
-        raise exc.HTTPBadGateway(str(e))
+        msg = str(e)
+        if e.response is not None:
+          msg += f' ; {e.response.text}'
+        logging.warning(msg)
+        logging.warning('\n'.join(traceback.format_tb(sys.exc_info()[2])))
+        abort(502, msg)
       raise
 
     if url != resp.url:
@@ -1472,7 +1587,7 @@ def requests_fn(fn):
 
     # check response size for text/ and application/ Content-Types
     type = resp.headers.get('Content-Type', '')
-    if type.startswith('text/') or type.startswith('application/'):
+    if type and (type.startswith('text/') or type.startswith('application/')):
       length = resp.headers.get('Content-Length')
       if is_int(length):
         length = int(length)
@@ -1487,6 +1602,8 @@ def requests_fn(fn):
         if gateway:
           resp.raise_for_status()
 
+    logging.info(f'Received {resp.status_code}')
+
     return resp
 
   return call
@@ -1509,7 +1626,7 @@ def requests_post_with_redirects(url, *args, **kwargs):
 
   Raises: TooManyRedirects
   """
-  for i in range(requests.models.DEFAULT_REDIRECT_LIMIT):
+  for _ in range(requests.models.DEFAULT_REDIRECT_LIMIT):
     resp = requests_post(url, *args, allow_redirects=False, **kwargs)
     url = resp.headers.get('Location')
     if resp.is_redirect and url:
@@ -1553,7 +1670,7 @@ def follow_redirects(url, **kwargs):
   """
   try:
     # default scheme to http
-    parsed = urllib.parse.urlparse(url)
+    parsed = urlparse(url)
     if not parsed.scheme:
       url = 'http://' + url
     resolved = requests_head(url, allow_redirects=True, **kwargs)
@@ -1651,12 +1768,12 @@ class UrlCanonicalizer(object):
     elif self.reject and self.reject.match(url):
       return None
 
-    parsed = urllib.parse.urlparse(url)
+    parsed = urlparse(url)
     domain = parsed.hostname
     if not domain:
       return None
-    elif self.domain and not (domain == self.domain or
-                              domain.endswith('.' + self.domain)):
+    elif (self.domain and domain != self.domain
+          and not domain.endswith('.' + self.domain)):
       return None
     if domain.startswith('www.'):
       domain = domain[4:]
@@ -1782,7 +1899,7 @@ def parse_html(input, **kwargs):
     # so, if charset isn't explicitly provided, pass on the raw bytes and let
     # BS4/UnicodeDammit figure it out from <meta charset> tag or anything else.
     # https://github.com/snarfed/granary/issues/171
-    content_type = input.headers.get('content-type', '')
+    content_type = input.headers.get('content-type') or ''
     input = input.text if 'charset' in content_type else input.content
 
   return bs4.BeautifulSoup(input, **kwargs)
